"""Customizable epochs for datetime/timedelta calculations.
"""
import numpy as np
cimport numpy as np
import pandas as pd

from pdcast.util.time cimport calendar
from pdcast.util.time import calendar
from pdcast.util.time cimport unit
from pdcast.util.time import unit
cimport pdcast.util.time.datetime as datetime_util
import pdcast.util.time.datetime as datetime_util
from pdcast.util.type_hints import datetime_like


#########################
####    CONSTANTS    ####
#########################


# aliases for commonly-encountered computing epochs
cdef dict epoch_aliases = {  # earliest - latest
    "julian":           np.datetime64("-4713-11-24 12:00:00"),
    "gregorian":        np.datetime64("1582-10-14 00:00:00"),
    "ntfs":             np.datetime64("1601-01-01 00:00:00"),
    "modified julian":  np.datetime64("1858-11-16 00:00:00"),
    "reduced julian":   np.datetime64("1858-11-16 12:00:00"),
    "lotus":            np.datetime64("1899-12-30 00:00:00"),
    "ntp":              np.datetime64("1900-01-01 00:00:00"),
    "risc":             np.datetime64("1900-01-01 00:00:00"),  # alias
    "labview":          np.datetime64("1904-01-01 00:00:00"),
    "sas":              np.datetime64("1960-01-01 00:00:00"),
    "utc":              np.datetime64("1970-01-01 00:00:00"),
    "unix":             np.datetime64("1970-01-01 00:00:00"),  # alias
    "posix":            np.datetime64("1970-01-01 00:00:00"),  # alias
    "fat":              np.datetime64("1980-01-01 00:00:00"),
    "zip":              np.datetime64("1980-01-01 00:00:00"),  # alias
    "gps":              np.datetime64("1980-01-06 00:00:00"),
    "j2000":            np.datetime64("2000-01-01 12:00:00"),
    "cocoa":            np.datetime64("2001-01-01 00:00:00")
}


epoch_aliases_public = epoch_aliases


######################
####    PUBLIC    ####
######################


cdef class Epoch:
    """Convert an epoch specifier into an integer nanosecond offset from the
    UTC epoch (1970-01-01 00:00:00).

    Parameters
    ----------
    arg : str | datetime-like
        The epoch specifier to use.  Can be one of the shorthand specifiers
        given in ``epoch_aliases`` or a direct datetime object.

    Returns
    -------
    Epoch
        An integer nanosecond offset from the UTC epoch.
    """

    def __init__(self, date: str | datetime_like):
        if isinstance(date, str):
            date = epoch_aliases[date.lower()]

        # separate calendar date components
        if isinstance(date, np.datetime64):
            Y, M, D = [date.astype(f"M8[{x}]") for x in "YMD"]
            self.year = Y.astype(np.int64) + 1970
            self.month = (M - Y).astype(np.uint8) + 1
            self.day = (D - M).astype(np.uint8) + 1
        else:  # duck type
            self.year = date.year
            self.month = date.month
            self.day = date.day

        # get nanosecond offset from UTC
        if isinstance(date, pd.Timestamp):
            self.offset = datetime_util.pandas_timestamp_to_ns(date)
        elif isinstance(date, np.datetime64):
            self.offset = datetime_util.numpy_datetime64_to_ns(date)
        else:  # duck type (datetime.datetime)
            self.offset = datetime_util.pydatetime_to_ns(date)

        # compute ordinal day
        ordinal = calendar.date_to_days(self.year, self.month, self.day)
        ordinal -= calendar.date_to_days(self.year, 1, 1)
        self.ordinal = ordinal + 1

    def __bool__(self) -> bool:
        return bool(self.offset)

    def __int__(self) -> int:
        return self.offset

    def __repr__(self) -> str:
        return f"{np.datetime64(self.offset // 10**9, 's')}"

    def __str__(self) -> str:
        return f"{np.datetime64(self.offset // 10**9, 's')}"

    def __eq__(self, other) -> bool:
        return isinstance(other, Epoch) and self.offset == other.offset

    def __lt__(self, other) -> bool:
        return isinstance(other, Epoch) and self.offset < other.offset

    def __gt__(self, other) -> bool:
        return isinstance(other, Epoch) and self.offset > other.offset



# pylint: disable=redefined-outer-name, unused-argument
import pandas as pd

from pdcast.decorators.attachable import attachable
from pdcast.decorators.dispatch import dispatch
from pdcast.decorators.extension import extension_func
from pdcast.util.numeric import real, imag, within_tol
from pdcast.util.round import Tolerance
from pdcast.util.type_hints import numeric

from .round import round


######################
####    PUBLIC    ####
######################


@attachable
@extension_func
@dispatch("series")
def snap(series: pd.Series, tol: numeric = 1e-6) -> pd.Series:
    """Snap each element of the series to the nearest integer if it is
    within the specified tolerance.
    """
    if not tol:
        return series.copy()

    rounded = round(series, rule="half_even")
    return series.where(~within_tol(series, rounded, tol=tol.real), rounded)


#########################
####    ARGUMENTS    ####
#########################


@snap.argument
def tol(val: numeric, state: dict) -> Tolerance:
    """The maximum amount of precision loss that can occur before an error
    is raised.

    Parameters
    ----------
    val : numeric
        A scalar numeric that is coercible to
        :class:`Decimal <python:decimal.Decimal>`.  In the case of complex
        values, their real and imaginary components are considered separately.
        Defaults to ``1e-6``

    Returns
    -------
    Tolerance
        A ``Tolerance`` object that consists of two
        :class:`Decimal <python:decimal.Decimal>` values, one for both the real
        and imaginary components.  This maintains the highest possible
        precision in both cases.

    Raises
    ------
    TypeError
        If ``val`` could not be coerced into a
        :class:`Decimal <python:decimal.Decimal>` representation.
    ValueError
        If the real or imaginary component of ``val`` is not positive.

    Notes
    -----
    Precision loss is defined using a 2-sided window around each of the
    observed values.  The size of this window is directly controlled by
    this argument.  If a conversion causes any value to be coerced outside
    this window, then a :class:`ValueError <python:ValueError>` will be raised.

    This argument only affects numeric conversions.

    Examples
    --------
    The input to this argument must be a positive numeric that is
    coercible to :class:`Decimal <python:decimal.Decimal>`.

    .. doctest::

        >>> cast(1.001, "int", tol=0.01)
        0    1
        dtype: int64
        >>> cast(1.001, "int", tol=0)
        Traceback (most recent call last):
            ...
        ValueError: precision loss exceeds tolerance 0 at index [0]

    If a complex value is given, then its real and imaginary components
    will be considered separately.

    .. doctest::

        >>> cast(1.001+0.001j, "int", tol=0.01+0.01j)
        0    1
        dtype: int64
        >>> cast(1.001+0.001j, "int", tol=0.01+0j)
        Traceback (most recent call last):
            ...
        ValueError: imaginary component exceeds tolerance 0 at index [0]

    This argument also has special behavior around the min/max of bounded
    numerics, like integers and booleans.  If a value would normally
    overflow, but falls within tolerance of these bounds, then it will be
    clipped to fit rather than raise an
    :class:`OverflowError <python:OverflowError>`.

    .. doctest::

        >>> cast(129, "int8", tol=2)
        0    127
        dtype: int8
        >>> cast(129, "int8", tol=0)
        Traceback (most recent call last):
            ...
        OverflowError: values exceed int8 range at index [0]

    Additionally, this argument controls the maximum amount of precision
    loss that can occur when
    :func:`downcasting <pdcast.convert.arguments.downcast>` numeric values.

    .. doctest::

        >>> cast(1.1, "float", tol=0, downcast=True)
        0    1.1
        dtype: float64
        >>> cast(1.1, "float", tol=0.001, downcast=True)
        0    1.099609
        dtype: float16

    Setting this to infinity ignores precision loss entirely.

    .. doctest::

        >>> cast(1.5, "int", tol=np.inf)
        0    2
        dtype: int64
        >>> cast(np.inf, "int64", tol=np.inf)
        0    9223372036854775807
        dtype: int64

    .. note::

        For integer conversions, this is equivalent to setting
        :func:`rounding <pdcast.convert.arguments.rounding>` to
        ``"half_even"``, with additional clipping around the minimum and
        maximum values.
    """
    if isinstance(val, Tolerance):
        return val
    return Tolerance(val)


@snap.overload("int")
def _snap_integer(series: pd.Series, tol: Tolerance) -> pd.Series:
    """Snap each element of the series to the nearest integer if it is
    within the specified tolerance.

    For integers, this is an identity function.
    """
    return series


@snap.overload("complex")
def _snap_complex(series: pd.Series, tol: Tolerance) -> pd.Series:
    """Snap each element of the series to the nearest integer if it is
    within the specified tolerance.
    """
    if not tol:  # trivial case, tol=0
        return series.copy()

    real_part = snap(real(series), tol=tol.real)
    imag_part = snap(imag(series), tol=tol.imag)
    return real_part + imag_part * 1j



"""This package implements the ``cast()`` function, its stand-alone
equivalents, and related tools for Series construction and manipulation.

Functions
---------
cast()
    An ``astype()`` equivalent for the ``pdcast`` type system, which allows for
    conversion between arbitrary data types.

to_boolean()
    Boolean-specific version of cast().

to_integer()
    Integer-specific version of cast().

to_float()
    Float-specific version of cast().

to_complex()
    Complex-specific version of cast().

to_decimal()
    Decimal-specific version of cast().

to_datetime()
    Datetime-specific version of cast().

to_timedelta()
    Timedelta-specific version of cast().

to_string()
    String-specific version of cast().
"""
from .arguments import *
from .base import *
from .boolean import *
from .integer import *
from .float import *
from .complex import *
from .decimal import *
from .datetime import *
from .timedelta import *
from .string import *
from .object import *
from .sparse import *
from .categorical import *



"""This module contains utility functions for manipulating timedelta-like data
in a variety of formats.

Functions
---------
pandas_timedelta_to_ns()
    Convert ``pandas.Timedelta`` objects into an integer number of nanoseconds.

pytimedelta_to_ns()
    Convert python ``datetime.timedelta`` objects into an integer number of
    nanoseconds.

numpy_timedelta64_to_ns()
    Convert ``numpy.timedelta64`` objects into an integer number of nanoseconds
    from the given epoch.

timedelta_string_to_ns()
    Convert a vector of timedelta strings into an integer number of nanoseconds
    from the given epoch.
"""
from cpython cimport datetime
import datetime
import decimal
import re

import numpy as np
cimport numpy as np
import pandas as pd

from pdcast.util.time cimport epoch
from pdcast.util.type_hints import datetime_like

from .unit cimport as_ns
from .unit import convert_unit, round_months_to_ns, round_years_to_ns


################################
####    PANDAS TIMEDELTA    ####
################################


cpdef inline object pandas_timedelta_to_ns(object delta):
    """Convert a pandas Timedelta into an integer number of nanoseconds."""
    # NOTE: as of pandas 2.0, Timedeltas can hold non-ns units

    if delta.unit == "ns":
        return delta.value
    return int(delta.asm8.view("i8")) * as_ns[delta.unit]


###############################
####    PYTHON DATETIME    ####
###############################


cpdef inline object pytimedelta_to_ns(datetime.timedelta delta):
    """Convert a python timedelta into an integer number of nanoseconds."""
    return (
        delta.days * as_ns["D"] +
        delta.seconds * as_ns["s"] +
        delta.microseconds * as_ns["us"]
    )


################################
####    NUMPY DATETIME64    ####
################################


cpdef inline object numpy_timedelta64_to_ns(
    object delta,
    epoch.Epoch since,
    str unit = None,
    long int step_size = -1
):
    """Convert a numpy timedelta64 into an integer number of nanoseconds."""
    cdef object result

    if unit is None or step_size < 0:
        unit, step_size = np.datetime_data(delta)

    result = int(delta.view(np.int64)) * step_size
    return convert_unit(result, unit, "ns", since=since)


######################
####    STRING    ####
######################


cdef dict timedelta_formats_regex():
    """Compile a list of regular expressions to capture and parse recognized
    timedelta strings.

    Matches both abbreviated ('1h22m', '1 hour, 22 minutes', etc.) and clock
    format ('01:22:00', '1:22', '00:01:22:00') strings, with precision up to
    years and months and down to nanoseconds.
    """
    # abbreviated capture groups.  NOTE: group names match unit.as_ns.
    cdef str Y = r"(?P<Y>[\d.]+)(?:ys?|yrs?.?|years?)"
    cdef str M = r"(?P<M>[\d.]+)(?:mos?.?|mths?.?|months?)"
    cdef str W = r"(?P<W>[\d.]+)(?:w|wks?|weeks?)"
    cdef str D = r"(?P<D>[\d.]+)(?:d|dys?|days?)"
    cdef str h = r"(?P<h>[\d.]+)(?:h|hrs?|hours?)"
    cdef str m = r"(?P<m>[\d.]+)(?:m|(mins?)|(minutes?))"
    cdef str s = r"(?P<s>[\d.]+)(?:s|secs?|seconds?)"
    cdef str ms = r"(?P<ms>[\d.]+)(?:ms|msecs?|millisecs?|milliseconds?)"
    cdef str us = r"(?P<us>[\d.]+)(?:us|usecs?|microsecs?|microseconds?)"
    cdef str ns = r"(?P<ns>[\d.]+)(?:ns|nsecs?|nanosecs?|nanoseconds?)"

    # clock format capture groups
    cdef str day_clock = (
        r"(?P<D>\d+):(?P<h>\d{2}):(?P<m>\d{2}):(?P<s>\d{2}(?:\.\d+)?)"
    )
    cdef str hour_clock = r"(?P<h>\d+):(?P<m>\d{2}):(?P<s>\d{2}(?:\.\d+)?)"
    cdef str minute_clock = r"(?P<m>\d{1,2}):(?P<s>\d{2}(?:\.\d+)?)"
    cdef str second_clock = r":(?P<s>\d{2}(?:\.\d+)?)"

    # mark optional groups with included separators
    opt = lambda x: rf"(?:{x}(?:[,/])?)?"

    # compiled timedelta formats
    return {
        "abbrev": re.compile(
            rf"^(?P<sign>[+-])?{opt(Y)}{opt(M)}{opt(W)}{opt(D)}{opt(h)}"
            rf"{opt(m)}{opt(s)}{opt(ms)}{opt(us)}{opt(ns)}$"
        ),
        "day": re.compile(rf"^(?P<sign>[+-])?{opt(W)}{opt(D)}{hour_clock}$"),
        "hour": re.compile(rf"^(?P<sign>[+-])?{day_clock}$"),
        "minute": re.compile(rf"^(?P<sign>[+-])?{minute_clock}$"),
        "second": re.compile(rf"^(?P<sign>[+-])?{second_clock}$")
    }


# compiled timedelta string regex
cdef dict timedelta_regex = timedelta_formats_regex()


cpdef object timedelta_string_to_ns(
    str input_string,
    bint as_hours,
    epoch.Epoch since
):
    """Convert a timedelta string into an integer number of nanoseconds."""
    cdef object match
    cdef char sign
    cdef str time_format
    cdef object regex
    cdef dict groups
    cdef object result
    cdef str k
    cdef str v

    # preprocess - remove whitespace and case
    input_string = re.sub(r"\s+", "", input_string).lower()

    # test all possible formats
    for time_format, regex in timedelta_regex.items():
        match = regex.match(input_string)
        if match and match.group():  # match found and not empty
            groups = match.groupdict()
            sign = -1 if groups["sign"] == "-" else 1

            # strings of the form '1:22' are ambiguous; do they represent
            # hours and minutes or minutes and seconds?  By default, we assume
            # the latter, but if `as_hours=True`, we reverse that assumption
            if as_hours and time_format == "minute":
                groups["h"] = groups["m"]
                groups["m"] = groups["s"]
                groups.pop("s")

            # build result
            result = 0
            for k, v in groups.items():
                if not v or k == "sign":
                    continue
                if k == "M":
                    result += int(round_months_to_ns(
                        decimal.Decimal(v),
                        since=since
                    ))
                elif k == "Y":
                    result += int(round_years_to_ns(
                        decimal.Decimal(v),
                        since=since
                    ))
                else:
                    result += int(as_ns[k] * decimal.Decimal(v))
            return sign * result

    # string could not be matched
    raise ValueError(f"could not parse timedelta string: {repr(input_string)}")



"""This module describes a mechanism for automatically generating pandas
``ExtensionDtype`` and ``ExtensionArray`` objects from an associated
``ScalarType``.

Classes
-------
ObjectDtype
    A pandas ``ExtensionDtype`` that delegates certain functionality to its
    associated ``ScalarType`` and comes with its own ``ExtensionArray``.

ObjectArray
    A pandas ``ExtensionArray`` that stores arbitrary objects in an
    explicitly-labeled ``dtype: object`` array.

Functions
---------
construct_object_dtype()
    A factory for ``ObjectDtype`` definitions.

construct_array_type()
    A factory for ``ObjectArray`` definitions.
"""
# pylint: disable=unused-argument
from __future__ import annotations

from collections import Counter
import numbers
import sys
from typing import Any, Iterable, Iterator, Literal, NoReturn

import numpy as np
import pandas as pd
from pandas.api.extensions import register_extension_dtype
from pandas.core.arrays import ExtensionArray, PandasArray
from pandas.core.dtypes.base import ExtensionDtype
from pandas.core.dtypes.generic import ABCDataFrame, ABCIndex, ABCSeries  # type: ignore

from bertrand.util.type_hints import array_like, dtype_like


# TODO: merge this with the equivalent stub


# TODO: run these through pandas test suite to see if anything is broken.
# -> need to implement _accumulate()


# TODO: use a classmethod to inject the ScalarType into the array definition rather
# than using a private attribute.  Maybe do the same with the corresponding dtype,
# although these have to stay unique for each ScalarType.  We can then just implement
# everything in the main definition and prevent lint errors.

# Now that we're in pure python, we can dynamically compute __doc__ strings based on
# the associated ScalarType.  This should be done for both the dtype and array


######################
####    PUBLIC    ####
######################


def construct_object_dtype(
    pdcast_type: Any,
    is_boolean: bool,
    is_numeric: bool,
    kind: valid_kinds = "O",
    nullable: bool = True,
    common_dtype: dtype_like | None = None
) -> ObjectDtype:
    """Construct a new pandas ``ExtensionDtype`` to refer to elements
    of this type.

    This method is only invoked if no explicit ``dtype`` is assigned to
    this ``ScalarType``.
    """
    # hack - allows us to use the same name for the global and local classes
    global ObjectDtype
    _ObjectDtype = ObjectDtype
    _kind = kind

    @register_extension_dtype
    class ObjectDtype(_ObjectDtype):

        _pdcast_type = pdcast_type
        name = str(pdcast_type)
        type = pdcast_type.type_def
        kind = _kind
        na_value = pdcast_type.na_value
        _is_boolean = is_boolean
        _is_numeric = is_numeric
        _can_hold_na = nullable

        @classmethod
        def construct_array_type(cls) -> type:
            """Build an ExtensionArray class to store objects of this dtype.

            This method automatically generates a new ObjectArray class
            specifically for this type.  These act like normal ``dtype: object``
            arrays, but are explicitly labeled with this dtype.
            """
            return construct_array_type(pdcast_type)

        def _get_common_dtype(self, dtypes: list) -> dtype_like | None:
            """Return the common dtype, if one exists.

            Used in `find_common_type` implementation. This is for example used
            to determine the resulting dtype in a concat operation.

            If no common dtype exists, return None (which gives the other
            dtypes the chance to determine a common dtype). If all dtypes in
            the list return None, then the common dtype will be "object" dtype
            (this means it is never needed to return "object" dtype from this
            method itself).

            Parameters
            ----------
            dtypes : list of dtypes
                The dtypes for which to determine a common dtype. This is a
                list of np.dtype or ExtensionDtype instances.

            Returns
            -------
            Common dtype (np.dtype or ExtensionDtype) or None
            """
            if len(set(dtypes)) == 1:  # only itself
                return self
            return common_dtype

    ObjectDtype.__doc__ = (
        f"An abstract data type, automatically generated to store\n "
        f"{pdcast_type.type_def} objects."
    )
    return ObjectDtype()


def construct_array_type(pdcast_type: Any) -> type[ObjectArray]:
    """Create a new ExtensionArray definition to store objects of the given
    type.
    """
    # hack - allows us to use the same name for the global and local classes
    global ObjectArray
    _ObjectArray = ObjectArray

    class ObjectArray(_ObjectArray):

        def __init__(self, *args, **kwargs):
            self._pdcast_type = pdcast_type
            super().__init__(*args, **kwargs)

    # replace docstring
    ObjectArray.__doc__ = (
        f"An abstract data type, automatically generated to store\n "
        f"{str(pdcast_type)} objects."
    )
    return ObjectArray


#######################
####    PRIVATE    ####
#######################


# pylint: disable=invalid-name
boolean_array = np.ndarray[np.bool_, np.dtype[np.bool_]]
valid_kinds = Literal['b', 'i', 'u', 'f', 'c', 'm', 'M', 'O', 'S', 'U', 'V']



class ObjectDtype(ExtensionDtype):
    """Base class for automatically-generated ExtensionDtype definitions.

    This class allows :class:`ScalarType` definitions that do not define an
    explicit ``.dtype`` field to automatically generate one according to
    :ref:`existing pandas guidelines <pandas:extending>`.  The resulting arrays
    are essentially identical to ``dtype: object`` arrays, but are explicitly
    labeled and have better integration with base pandas.  They may also be
    slightly more performant in some cases.
    """

    def __init__(self):
        # require use of construct_object_dtype() factory
        if not hasattr(self, "_pdcast_type"):
            raise NotImplementedError(
                "ObjectDtype must have an associated ScalarType"
            )

    @classmethod
    def construct_from_string(cls, string: str) -> NoReturn:
        """Construct this type from a string.

        This is useful mainly for data types that accept parameters.
        For example, a period dtype accepts a frequency parameter that
        can be set as ``period[H]`` (where H means hourly frequency).
        By default, in the abstract class, just the name of the type is
        expected. But subclasses can overwrite this method to accept
        parameters.

        Parameters
        ----------
        string : str
            The name of the type, for example ``category``.

        Returns
        -------
        ExtensionDtype
            Instance of the dtype.

        Raises
        ------
        TypeError
            If a class cannot be constructed from this 'string'.

        Examples
        --------
        For extension dtypes with arguments the following may be an
        adequate implementation.

        >>> import re
        >>> @classmethod
        ... def construct_from_string(cls, string):
        ...     pattern = re.compile(r"^my_type\[(?P<arg_name>.+)\]$")
        ...     match = pattern.match(string)
        ...     if match:
        ...         return cls(**match.groupdict())
        ...     else:
        ...         raise TypeError(
        ...             f"Cannot construct a '{cls.__name__}' from "
        ...             f"'{string}'"
        ...         )
        """
        # disable string construction - use ``pdcast.resolve_type()`` instead
        raise TypeError(f"Cannot construct a '{cls.__name__}' from '{string}'")

    def __getattr__(self, name: str) -> Any:
        """Delegate attribute lookups to the underlying ScalarType."""
        try:
            return self._pdcast_type.kwargs[name]
        except KeyError as err:
            err_msg = (
                f"{repr(type(self).__name__)} object has no attribute: "
                f"{repr(name)}"
            )
            raise AttributeError(err_msg) from err

    def __dir__(self) -> list:
        # direct ObjectDtype attributes
        result = dir(type(self))
        result += list(self.__dict__.keys())

        # ScalarType kwargs
        result += [x for x in self._pdcast_type.kwargs if x not in result]
        return result

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self._pdcast_type)})"


class ObjectArray(ExtensionArray):
    """Base class for automatically-generated ExtensionArray definitions.

    This class allows :class:`ScalarType` definitions that do not define an
    explicit ``.dtype`` field to automatically generate one according to the
    :ref:`existing pandas guidelines <pandas:extending>`.  The resulting arrays
    are essentially identical to ``dtype: object`` arrays, but are explicitly
    labeled and have better integration with base pandas.  They may also be
    slightly more performant in some cases.
    """

    __array_priority__: int = 1000  # this is used in pandas test code
    data: np.ndarray[Any, np.dtype[Any]]
    _data: np.ndarray[Any, np.dtype[Any]]
    _items: np.ndarray[Any, np.dtype[Any]]

    def __init__(
        self,
        values: Iterable[Any],
        dtype: dtype_like | None = None,
        copy: bool = False,
    ):
        # NOTE: this does NOT coerce inputs; that's handled by cast() itself

        # aliases for common attributes to ensure pandas support
        self.data = self._data = self._items = np.asarray(values, dtype=object)

        # require use of construct_array_type() factory
        if not hasattr(self, "_pdcast_type"):
            raise NotImplementedError(
                "ObjectArray must have an associated ScalarType"
            )

    ############################
    ####    CONSTRUCTORS    ####
    ############################

    @classmethod
    def _from_sequence(
        cls,
        scalars: Iterable[Any],
        dtype: dtype_like | None = None,
        copy: bool = False
    ) -> ObjectArray:
        """Construct a new ExtensionArray from a sequence of scalars.

        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type`` or be converted into this type in this
            method.
        dtype : dtype, optional
            Construct for this particular dtype.  This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.

        Returns
        -------
        ExtensionArray

        Notes
        -----
        This **does not** coerce results to the specified type.  For
        conversions, see the docs on :func:`cast` and its related functions.
        All this does is wrap the scalars in an object array.
        """
        return cls(scalars, dtype=dtype, copy=copy)

    @classmethod
    def _from_factorized(
        cls, values: np.ndarray[np.int64, np.dtype[np.int64]], original: ExtensionArray
    ) -> ObjectArray:
        """Reconstruct an ExtensionArray after factorization.

        Parameters
        ----------
        values : ndarray
            An integer ndarray with the factorized values.
        original : ExtensionArray
            The original ExtensionArray that factorize was called on.

        See Also
        --------
        factorize : Top-level factorize method that dispatches here.
        ExtensionArray.factorize : Encode the array as an enumerated type.
        """
        return cls(values, dtype=original.dtype)

    @classmethod
    def _concat_same_type(cls, to_concat: Iterable[ExtensionArray]) -> ObjectArray:
        """Concatenate multiple array of this dtype.

        Parameters
        ----------
        to_concat : Iterable[ExtensionArray]
            sequence of this type

        Returns
        -------
        ExtensionArray
        """
        return cls(np.concatenate([x.data for x in to_concat]))

    ##########################
    ####    ATTRIBUTES    ####
    ##########################

    @property
    def dtype(self):
        """An instance of 'ExtensionDtype'."""
        return self._pdcast_type.dtype

    @property
    def _dtype(self):
        """An alias for ``self.dtype`` that ensures pandas compatibility."""
        return self.dtype

    @property
    def nbytes(self):
        """The number of bytes needed to store this object in memory."""
        n = len(self)
        if n:
            return n * sys.getsizeof(self[0])
        return 0

    ########################
    ####    INTERNAL    ####
    ########################

    def _values_for_argsort(self) -> np.ndarray:
        """Return values for sorting.

        Returns
        -------
        ndarray
            The transformed values should maintain the ordering between values
            within the array.

        See Also
        --------
        ExtensionArray.argsort : Return the indices that would sort this array.

        Notes
        -----
        The caller is responsible for *not* modifying these values in-place, so
        it is safe for implementors to give views on `self`.

        Functions that use this (e.g. ExtensionArray.argsort) should ignore
        entries with missing values in the original array (according to
        `self.isna()`). This means that the corresponding entries in the
        returned array don't need to be modified to sort correctly.
        """
        # Note: this is used in `ExtensionArray.argsort/argmin/argmax`.
        return self.data

    def _values_for_factorize(self) -> tuple[np.ndarray[Any, np.dtype[Any]], Any]:
        """Return an array and missing value suitable for factorization.

        Returns
        -------
        values : ndarray
            An array suitable for factorization. This should maintain order
            and be a supported dtype (Float64, Int64, UInt64, String, Object).
            By default, the extension array is cast to object dtype.
        na_value : object
            The value in `values` to consider missing. This will be treated
            as NA in the factorization routines, so it will be coded as
            `na_sentinel` and not included in `uniques`.

        Notes
        -----
        The values returned by this method are also used in
        :func:`pandas.util.hash_pandas_object`.
        """
        return self.data, self.dtype.na_value

    def _reduce(self, name: str, skipna: bool = True, **kwargs: Any) -> Any:
        """Return a scalar result of performing the reduction operation.

        Parameters
        ----------
        name : str
            Name of the function, supported values are:
            { any, all, min, max, sum, mean, median, prod,
            std, var, sem, kurt, skew }.
        skipna : bool, default True
            If True, skip NaN values.
        **kwargs
            Additional keyword arguments passed to the reduction function.
            Currently, `ddof` is the only supported kwarg.

        Returns
        -------
        scalar

        Raises
        ------
        TypeError : subclass does not define reductions
        """
        # NOTE: this implementation is taken from pandas test suite

        if skipna:
            # If we don't have any NAs, we can ignore skipna
            if self.isna().any():
                other = self[~self.isna()]
                return other._reduce(name, **kwargs)

        if name == "sum" and len(self) == 0:
            # GH#29630 avoid returning int 0 or np.bool_(False) on old numpy
            item_type = self._pdcast_type.type_def
            return item_type(0)

        try:
            op = getattr(self.data, name)
        except AttributeError:
            raise NotImplementedError(
                f"{str(self._pdcast_type)} does not support the {name} "
                f"operation"
            )
        return op(axis=0)

    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs: Any, **kwargs: Any
    ) -> ObjectArray:
        """Allows ObjectArrays to utilize numpy ufuncs natively, without
        being coerced to ``dtype: object``.

        This implementation is adapted from the
        :class:`example <numpy.lib.mixins.NDArrayOperatorsMixin>` given by
        numpy.  It might require further modifications.
        """
        # pandas unboxes these so we don't need to implement them ourselves
        if any(
            isinstance(other, (ABCSeries, ABCIndex, ABCDataFrame))  # type: ignore
            for other in inputs
        ):
            return NotImplemented

        # get handled values
        array_like = (np.ndarray, ExtensionArray)
        scalar_like = (self._pdcast_type.type_def,)
        if self.dtype._is_numeric:
            scalar_like += (numbers.Number,)

        if not all(isinstance(t, array_like + scalar_like) for t in inputs):
            return NotImplemented

        inputs = tuple(x.data if isinstance(x, ExtensionArray) else x for x in inputs)
        result = getattr(ufunc, method)(*inputs, **kwargs)

        def reconstruct(x):
            if isinstance(x, scalar_like):
                return x
            return type(self)._from_sequence(x)

        if isinstance(result, tuple):
            return tuple(reconstruct(x) for x in result)
        return reconstruct(result)

    def __arrow_array__(self, type: dtype_like | None = None) -> Any:
        """Convert the underlying array values into a pyarrow Array.

        This method ensures that pyarrow knowns how to convert the specific
        extension array into a pyarrow.Array (also when included as a column in
        a pandas DataFrame).
        """
        import pyarrow

        return pyarrow.array(self.data, type=type)

    def __from_arrow__(self, array: Any) -> ObjectArray:
        """Convert a pyarrow array into a pandas ExtensionArray.

        This method controls the conversion back from pyarrow to a pandas
        ExtensionArray.  It receives a pyarrow Array or ChunkedArray as its
        only argument and is expected to return the appropriate pandas
        ExtensionArray for this dtype and the passed values
        """
        return self._from_sequence(array)

    ######################
    ####    PUBLIC    ####
    ######################

    def astype(self, dtype: dtype_like, copy: bool = True) -> array_like:
        """Cast to a NumPy array or ExtensionArray with 'dtype'.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the array is cast.
        copy : bool, default True
            Whether to copy the data, even if not necessary. If False, a copy
            is made only if the old dtype does not match the new dtype.

        Returns
        -------
        array : np.ndarray or ExtensionArray
            An ExtensionArray if dtype is ExtensionDtype, otherwise a NumPy
            ndarray with 'dtype' for its dtype.
        """
        dtype = pd.api.types.pandas_dtype(dtype)
        if pd.core.dtypes.common.is_dtype_equal(dtype, self.dtype):
            if not copy:
                return self
            return self.copy()

        if isinstance(dtype, ExtensionDtype):
            cls = dtype.construct_array_type()
            return cls._from_sequence(self.data, dtype=dtype, copy=copy)

        return self.data.astype(dtype=dtype, copy=copy)

    def isna(self) -> boolean_array:
        """Detect missing values from the array."""
        return pd.isna(self.data)

    def take(
        self,
        indices: Iterable[int],
        allow_fill: bool = False,
        fill_value: Any | None = None
    ) -> ObjectArray:
        """Take elements from an array.

        Parameters
        ----------
        indices : sequence of int or one-dimensional np.ndarray of int
            Indices to be taken.
        allow_fill : bool, default False
            How to handle negative values in `indices`.
            * False: negative values in `indices` indicate positional indices
              from the right (the default). This is similar to
              :func:`numpy.take`.
            * True: negative values in `indices` indicate
              missing values. These values are set to `fill_value`. Any other
              other negative values raise a ``ValueError``.
        fill_value : any, optional
            Fill value to use for NA-indices when `allow_fill` is True.
            This may be ``None``, in which case the default NA value for
            the type, ``self.dtype.na_value``, is used.
            For many ExtensionArrays, there will be two representations of
            `fill_value`: a user-facing "boxed" scalar, and a low-level
            physical NA value. `fill_value` should be the user-facing version,
            and the implementation should handle translating that to the
            physical version for processing the take if necessary.

        Returns
        -------
        ExtensionArray

        Raises
        ------
        IndexError
            When the indices are out of bounds for the array.
        ValueError
            When `indices` contains negative values other than ``-1``
            and `allow_fill` is True.
    
        See Also
        --------
        numpy.take : Take elements from an array along an axis.
        api.extensions.take : Take elements from an array.

        Notes
        -----
        ExtensionArray.take is called by ``Series.__getitem__``, ``.loc``,
        ``iloc``, when `indices` is a sequence of values. Additionally,
        it's called by :meth:`Series.reindex`, or any other method
        that causes realignment, with a `fill_value`.

        Examples
        --------
        Here's an example implementation, which relies on casting the
        extension array to object dtype. This uses the helper method
        :func:`pandas.api.extensions.take`.

        .. code-block:: python

           def take(self, indices, allow_fill=False, fill_value=None):
               from pandas.core.algorithms import take

               # If the ExtensionArray is backed by an ndarray, then
               # just pass that here instead of coercing to object.
               data = self.astype(object)

               if allow_fill and fill_value is None:
                   fill_value = self.dtype.na_value

               # fill value should always be translated from the scalar
               # type for the array, to the physical storage type for
               # the data, before passing to take.

               result = take(data, indices, fill_value=fill_value,
                             allow_fill=allow_fill)
               return self._from_sequence(result, dtype=self.dtype)
        """
        from pandas.core.algorithms import take

        data = self.data
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        result = take(
            data,
            indices,
            fill_value=fill_value,
            allow_fill=allow_fill
        )
        return self._from_sequence(result, dtype=self.dtype)

    def copy(self) -> ObjectArray:
        """Return a copy of the array.

        Returns
        -------
        ExtensionArray
        """
        return type(self)(self.data.copy())

    ###########################
    ####    NEW METHODS    ####
    ###########################

    # TODO: implement unique()?  Is this necessary?

    # NOTE: these are not defined in the base ExtensionArray class, but are
    # called in several pandas operations.

    def round(self, *args: Any, **kwargs: Any) -> ObjectArray:
        """Default round implementation.  This simply passes through to
        np.round().
        """
        return self._from_sequence(self.data.round(*args, **kwargs))

    def value_counts(self, dropna: bool = True) -> pd.Series:
        """Returns a Series containing counts of each unique value.

        Parameters
        ----------
        dropna : bool, default True
            Don't include counts of missing values.

        Returns
        -------
        counts : Series

        See Also
        --------
        Series.value_counts
        """
        # compute counts without nans
        mask = self.isna()
        counts = Counter(self.data[~mask])
        result = pd.Series(
            counts.values(),
            index=pd.Index(list(counts.keys()), dtype=self.dtype)
        )
        if dropna:
            return result.astype("Int64")

        # if we want to include nans, count mask
        nans = pd.Series(
            [mask.sum()],
            index=pd.Index([self.dtype.na_value], dtype=self.dtype)
        )
        result = pd.concat([result, nans])
        return result.astype("Int64")

    ###############################
    ####    UNARY OPERATORS    ####
    ###############################

    def __abs__(self) -> ObjectArray:
        """Implement the :func:`abs() <python:abs>` function for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        return rectify(abs(self.data), self._pdcast_type)

    def __neg__(self) -> ObjectArray:
        """Implement the unary negative operator ``-`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        if self.dtype._is_boolean:
            return self.__invert__()

        return rectify(-self.data, self._pdcast_type)

    def __pos__(self) -> ObjectArray:
        """Implement the unary positive operator ``+`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        return rectify(+self.data, self._pdcast_type)

    def __invert__(self) -> ObjectArray:
        """Implement the unary inversion operator ``~`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        if self.dtype._is_boolean:
            return self.__xor__(True)

        return rectify(~self.data, self._pdcast_type)

    ################################
    ####    BINARY OPERATORS    ####
    ################################

    def __and__(self, other: Any) -> ExtensionArray:
        """Implement the binary AND operator ``&`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data & np.asarray(other), self._pdcast_type)

    def __or__(self, other: Any) -> ExtensionArray:
        """Implement the binary OR operator ``|`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data & np.asarray(other), self._pdcast_type)

    def __xor__(self, other: Any) -> ExtensionArray:
        """Implement the binary XOR operator ``^`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data & np.asarray(other), self._pdcast_type)

    def __rshift__(self, other: Any) -> ExtensionArray:
        """Implement the binary right-shift operator ``>>`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data & np.asarray(other), self._pdcast_type)

    def __lshift__(self, other: Any) -> ExtensionArray:
        """Implement the binary left-shift operator ``<<`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data & np.asarray(other), self._pdcast_type)

    ########################################
    ####    REVERSE BINARY OPERATORS    ####
    ########################################

    def __rand__(self, other: Any) -> ExtensionArray:
        """Implement the reverse binary AND operator ``&`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) & self.data, self._pdcast_type)

    def __ror__(self, other: Any) -> ExtensionArray:
        """Implement the reverse binary OR operator ``|`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) | self.data, self._pdcast_type)

    def __rxor__(self, other: Any) -> ExtensionArray:
        """Implement the reverse binary XOR operator ``^`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) ^ self.data, self._pdcast_type)

    def __rrshift__(self, other: Any) -> ExtensionArray:
        """Implement the reverse binary right-shift operator ``>>`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) >> self.data, self._pdcast_type)

    def __rlshift__(self, other: Any) -> ExtensionArray:
        """Implement the reverse binary left-shift operator ``<<`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) << self.data, self._pdcast_type)

    #########################################
    ####    IN-PLACE BINARY OPERATORS    ####
    #########################################

    def __iand__(self, other: Any) -> ExtensionArray:
        """Implement the in-place binary AND operator ``&=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data &= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __ior__(self, other: Any) -> ExtensionArray:
        """Implement the in-place binary OR operator ``|=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data |= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __ixor__(self, other: Any) -> ExtensionArray:
        """Implement the in-place binary XOR operator ``^=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data ^= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __irshift__(self, other: Any) -> ExtensionArray:
        """Implement the in-place binary right-shift operator ``>>=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data >>= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __ilshift__(self, other: Any) -> ExtensionArray:
        """Implement the in-place binary left-shift operator ``<<=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data <<= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    ##############################
    ####    MATH OPERATORS    ####
    ##############################

    # NOTE: math operators can change an array's type in unexpected ways, so
    # we always follow up with a detect_type call to rectify the result.

    def __add__(self, other: Any) -> ExtensionArray:
        """Implement the addition operator ``+`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data + np.asarray(other), self._pdcast_type)

    def __sub__(self, other: Any) -> ExtensionArray:
        """Implement the subtraction operator ``-`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data - np.asarray(other), self._pdcast_type)

    def __mul__(self, other: Any) -> ExtensionArray:
        """Implement the multiplication operator ``*`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data * np.asarray(other), self._pdcast_type)

    def __matmul__(self, other: Any) -> ExtensionArray:
        """Implement the matrix multiplication operator ``@`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data @ np.asarray(other), self._pdcast_type)

    def __truediv__(self, other: Any) -> ExtensionArray:
        """Implement the true division operator ``/`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data / np.asarray(other), self._pdcast_type)

    def __floordiv__(self, other: Any) -> ExtensionArray:
        """Implement the floor division operator ``//`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data // np.asarray(other), self._pdcast_type)

    def __mod__(self, other: Any) -> ExtensionArray:
        """Implement the modulo operator ``%`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(self.data % np.asarray(other), self._pdcast_type)

    def __divmod__(self, other: Any) -> ExtensionArray:
        """Implement the :func:`divmod() <python:divmod>` function for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(
            divmod(self.data, np.asarray(other)),
            self._pdcast_type
        )

    def __pow__(self, other: Any, mod: Any | None = None) -> ExtensionArray:
        """Implement the exponentiation operator ``**`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(
            pow(self.data, np.asarray(other), mod),
            self._pdcast_type
        )

    ######################################
    ####    REVERSE MATH OPERATORS    ####
    ######################################

    def __radd__(self, other: Any) -> ExtensionArray:
        """Implement the reverse addition operator ``+`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) + self.data, self._pdcast_type)

    def __rsub__(self, other: Any) -> ExtensionArray:
        """Implement the reverse subtraction operator ``-`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) - self.data, self._pdcast_type)

    def __rmul__(self, other: Any) -> ExtensionArray:
        """Implement the reverse multiplication operator ``*`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) * self.data, self._pdcast_type)

    def __rmatmul__(self, other: Any) -> ExtensionArray:
        """Implement the reverse matrix multiplication operator ``@`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) @ self.data, self._pdcast_type)

    def __rtruediv__(self, other: Any) -> ExtensionArray:
        """Implement the reverse true division operator ``/`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) / self.data, self._pdcast_type)

    def __rfloordiv__(self, other: Any) -> ExtensionArray:
        """Implement the reverse floor division operator ``//`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) // self.data, self._pdcast_type)

    def __rmod__(self, other: Any) -> ExtensionArray:
        """Implement the reverse modulo operator ``%`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(np.asarray(other) % self.data, self._pdcast_type)

    def __rdivmod__(self, other: Any) -> ExtensionArray:
        """Implement the reverse :func:`divmod() <python:divmod>` function for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(
            divmod(np.asarray(other), self.data),
            self._pdcast_type
        )

    def __rpow__(self, other: Any, mod: Any | None = None) -> ExtensionArray:
        """Implement the reverse exponentiation operator ``**`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return rectify(
            pow(np.asarray(other), self.data, mod),
            self._pdcast_type
        )

    #######################################
    ####    IN-PLACE MATH OPERATORS    ####
    #######################################

    def __iadd__(self, other: Any) -> ExtensionArray:
        """Implement the in-place addition operator ``+=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data += np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __isub__(self, other: Any) -> ExtensionArray:
        """Implement the in-place subtraction operator ``-=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data -= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __imul__(self, other: Any) -> ExtensionArray:
        """Implement the in-place multiplication operator ``*=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data *= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __imatmul__(self, other: Any) -> ExtensionArray:
        """Implement the in-place matrix multiplication operator ``@=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data @= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __itruediv__(self, other: Any) -> ExtensionArray:
        """Implement the in-place true division operator ``/=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data /= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __ifloordiv__(self, other: Any) -> ExtensionArray:
        """Implement the in-place floor division operator ``//=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data //= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __imod__(self, other: Any) -> ExtensionArray:
        """Implement the in-place modulo operator ``%=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data %= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    def __ipow__(self, other: Any, mod: Any | None = None) -> ExtensionArray:
        """Implement the in-place exponentiation operator ``**=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        self.data **= np.asarray(other)
        return rectify(self.data, self._pdcast_type)

    ####################################
    ####    COMPARISON OPERATORS    ####
    ####################################

    def __lt__(self, other: Any) -> boolean_array:
        """Implement the less-than operator ``<`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data < other

    def __le__(self, other: Any) -> boolean_array:
        """Implement the less-than-or-equal-to operator ``<=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data <= other

    def __eq__(self, other: Any) -> boolean_array:  # type: ignore
        """Implement the equality operator ``==`` (elementwise) for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data == other

    def __ne__(self, other: Any) -> boolean_array:  # type: ignore
        """Implement the inequality operator ``!=`` (elementwise) for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data != other

    def __gt__(self, other: Any) -> boolean_array:
        """Implement the greater-than operator ``>`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data > other

    def __ge__(self, other: Any) -> boolean_array:
        """Implement the greater-than-or-equal-to operator ``>=`` for
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        # NOTE: pandas recommends that we not handle these cases ourselves
        if isinstance(other, (ABCSeries, ABCIndex)):  # type: ignore
            return NotImplemented

        return self.data >= other

    #################################
    ####    CONTAINER METHODS    ####
    #################################

    def __getitem__(self, key: Any) -> Any | ObjectArray:
        """Select a subset of self.

        Parameters
        ----------
        key : int, slice, or ndarray
            * int: The position in 'self' to get.
            * slice: A slice object, where 'start', 'stop', and 'step' are
              integers or None
            * ndarray: A 1-d boolean NumPy ndarray the same length as 'self'
            * list[int]:  A list of int

        Returns
        -------
        scalar or ExtensionArray

        Notes
        -----
        For scalar ``key``, return a scalar value suitable for the array's
        type. This should be an instance of ``self.dtype.type``.

        For slice ``key``, return an instance of ``ExtensionArray``, even
        if the slice is length 0 or 1.

        For a boolean mask, return an instance of ``ExtensionArray``, filtered
        to the values where ``key`` is True.
        """
        if isinstance(key, numbers.Integral):
            return self.data[key]
        return self._from_sequence(self.data[key])

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set one or more values inplace.

        This method is not required to satisfy the pandas extension array
        interface.

        Parameters
        ----------
        key : int, ndarray, or slice
            When called from, e.g. ``Series.__setitem__``, ``key`` will be
            one of
            * scalar int
            * ndarray of integers.
            * boolean ndarray
            * slice object
        value : ExtensionDtype.type, Sequence[ExtensionDtype.type], or object
            value or values to be set of ``key``.
        """
        from pdcast import convert

        if pd.api.types.is_list_like(value):
            if pd.api.types.is_scalar(key):
                raise ValueError(
                    "setting an array element with a sequence."
                )
            value = np.asarray(convert.cast(value, self._pdcast_type))
        else:
            value = convert.cast(value, self._pdcast_type)[0]

        self.data[key] = value

    def __delitem__(self, key: Any) -> None:
        """Delete one or more values inplace.

        This always raises an error since numpy does not support deleting
        elements from an array.
        """
        del self.data[key]

    def __iter__(self) -> Iterator[Any]:
        """Iterate over elements of the array."""
        return iter(self.data)

    def __len__(self) -> int:
        """Length of this array

        Returns
        -------
        length : int
        """
        return len(self.data)

    def __contains__(self, other: Any) -> bool:
        """Implement the ``in`` keyword for membership tests on
        :class:`ObjectArrays <pdcast.ObjectArray>`.
        """
        return other in self.data


def rectify(arr: np.ndarray[Any, np.dtype[Any]], orig_dtype) -> ExtensionArray:
    """Rectify the result of a math operation to the detected type of the
    first element.

    Parameters
    ----------
    arr : np.ndarray
        The array to rectify.  This is typically the result of a math
        operation, which may or may not coerce the result to a different type.

    Returns
    -------
    ExtensionArray
        The rectified array.  This is always returned as a pandas-compatible
        `ExtensionArray` object.

    Notes
    -----
    The output type is determined by the first non-missing element of the
    input array.
    """
    from pdcast.detect import detect_type

    # get first non-missing element and detect its type
    is_na = pd.isna(arr)
    first = next(iter(arr[~is_na]), None)
    detected = detect_type(first)

    # check for empty array
    if detected is None:  # TODO: replace with NullType
        return arr

    # fill missing values with new type's na_value
    if detected != orig_dtype:
        arr[is_na] = detected.na_value

    # get pandas array type
    dtype = detected.dtype
    if isinstance(dtype, ExtensionDtype):
        array_type = dtype.construct_array_type()
    else:
        array_type = PandasArray

    # construct output array
    return array_type._from_sequence(arr, dtype=dtype, copy=False)



"""This module describes a base class for decorators that allow cooperative
attribute access with nested objects.
"""
from __future__ import annotations
from functools import update_wrapper, WRAPPER_ASSIGNMENTS
import inspect
from types import MappingProxyType
from typing import Any, Callable, Mapping


# shortcut for inspect.Parameter.empty
EMPTY: type = inspect.Parameter.empty


# shortcut for inspect.Parameter.kind
KINDS: dict[str, inspect._ParameterKind] = {
    "POSITIONAL_ONLY": inspect.Parameter.POSITIONAL_ONLY,
    "POSITIONAL_OR_KEYWORD": inspect.Parameter.POSITIONAL_OR_KEYWORD,
    "VAR_POSITIONAL": inspect.Parameter.VAR_POSITIONAL,
    "KEYWORD_ONLY": inspect.Parameter.KEYWORD_ONLY,
    "VAR_KEYWORD": inspect.Parameter.VAR_KEYWORD,
}


class FunctionDecorator:
    """Base class for cooperative decorators.

    These decorators implement the Gang of Four's `Decorator pattern
    <https://en.wikipedia.org/wiki/Decorator_pattern>`_ to  cooperatively pass
    attribute access down a (possibly nested) decorator stack.

    Parameters
    ----------
    func : Callable
        The function (or other callable) to decorate.  Any attributes that are
        not found on this decorator will be dynamically passed down to this
        object.
    **kwargs : dict
        Keyword arguments to use for multiple inheritance.

    Notes
    -----
    This expects all subclasses to define the following attribute at the class
    level:

        *   ``_reserved``: a set of strings describing attribute names that are
            reserved for this decorator.

    The special string ``"__wrapped__"`` is always added to this set,
    representing the decorated callable itself.  This is automatically assigned
    by :func:`update_wrapper <python:functools.update_wrapper>` during
    ``__init__``.
    """

    _reserved = set(WRAPPER_ASSIGNMENTS) | {"__wrapped__"}

    def __init__(self, func: Callable, **kwargs):
        super().__init__(**kwargs)  # allow multiple inheritance
        update_wrapper(self, func)  # sets __wrapped__, __name__, __doc__, etc.

    ################################
    ####    ATTRIBUTE ACCESS    ####
    ################################

    def __getattr__(self, name: str) -> Any:
        """Delegate getters to wrapped object."""
        return getattr(self.__wrapped__, name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Delegate setters to wrapped object."""
        if name in self._reserved or hasattr(type(self), name):
            super().__setattr__(name, value)
        else:
            setattr(self.__wrapped__, name, value)

    def __delattr__(self, name: str) -> None:
        """Delegate deleters to wrapped object."""
        # NOTE: hasattr() doesn't respect nested decorators
        try:
            object.__getattribute__(self, name)
            super().__delattr__(name)
        except AttributeError:
            delattr(self.__wrapped__, name)

    ################################
    ####     SPECIAL METHODS    ####
    ################################

    def __call__(self, *args, **kwargs):
        """Invoke the wrapped object's ``__call__()`` method."""
        return self.__wrapped__(*args, **kwargs)

    def __getitem__(self, key) -> Any:
        """Pass indexing to wrapped object."""
        return self.__wrapped__.__getitem__(key)

    def __setitem__(self, key, value) -> None:
        """Pass index assignment to wrapped object."""
        return self.__wrapped__.__setitem__(key, value)

    def __delitem__(self, key) -> None:
        return self.__wrapped__.__delitem__(key)

    def __contains__(self, item: Any) -> bool:
        """Pass ``in`` keyword to wrapped object."""
        return item in self.__wrapped__

    def __dir__(self) -> list:
        """Include attributes of wrapped object."""
        result = dir(type(self))
        result += [k for k in self.__dict__ if k not in result]
        result += [k for k in dir(self.__wrapped__) if k not in result]
        return result

    def __str__(self) -> str:
        """Pass ``str()`` calls to wrapped object."""
        return str(self.__wrapped__)

    def __repr__(self) -> str:
        """Pass ``repr()`` calls to wrapped object."""
        return repr(self.__wrapped__)


class Signature:
    """An extensible wrapper around an
    :class:`inspect.Parameter <python:inspect.Parameter>` object that allows
    easy modification of parameters and annotations.

    Parameters
    ----------
    func : Callable
        A function or other callable to introspect.
    """

    def __init__(self, func: Callable):
        self.func_name: str = func.__qualname__
        self.sig: inspect.Signature = inspect.signature(func)
        self.original: inspect.Signature = self.sig  # remember original sig

    @property
    def parameter_map(self) -> Mapping[str, inspect.Parameter]:
        """A read-only mapping from argument names to their corresponding
        :class:`Parameters <python:inspect.Parameter>`.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig
            <Signature (bar, baz=2, **kwargs)>
            >>> sig.parameter_map
            mappingproxy(OrderedDict([('bar', <Parameter "bar">), ('baz', <Parameter "baz=2">), ('kwargs', <Parameter "**kwargs">)]))
        """
        return self.sig.parameters

    @property
    def parameters(self) -> tuple[inspect.Parameter, ...]:
        """Access the :class:`Parameter <python:inspect.Parameter>` objects
        that are stored in a signature.

        Returns
        -------
        tuple[inspect.Parameter]
            An ordered tuple containing every parameter in the signature.

        Notes
        -----
        This property can be written to in order to allow parameters to be
        replaced, removed, and/or extended at runtime.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.parameters
            (<Parameter "bar">, <Parameter "baz=2">, <Parameter "**kwargs">)
            >>> sig.parameters = [par.replace(annotation=int) for par in sig.parameters]
            >>> sig.parameters
            (<Parameter "bar: int">, <Parameter "baz: int = 2">, <Parameter "**kwargs: int">)
            >>> del sig.parameters
            >>> sig.parameters
            (<Parameter "bar">, <Parameter "baz=2">, <Parameter "**kwargs">)
        """
        return tuple(self.sig.parameters.values())

    @parameters.setter
    def parameters(self, val: tuple[inspect.Parameter]) -> None:
        self.sig = self.sig.replace(parameters=val)

    @parameters.deleter
    def parameters(self) -> None:
        self.sig = self.original

    @property
    def return_annotation(self) -> Any:
        """Access the return annotation of the signature.

        Returns
        -------
        Any
            The raw return annotation for the signature.

        Notes
        -----
        This property can be written to in order to allow the return annotation
        to be replaced or removed at runtime.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.return_annotation
            <class 'inspect._empty'>
            >>> sig.return_annotation = dict[str, int]
            >>> sig.return_annotation
            dict[str, int]
            >>> sig
            '(bar, baz= 2, **kwargs) -> dict[str, int]'
            >>> del sig.return_annotation
            >>> sig.return_annotation
            <class 'inspect._empty'>
        """
        return self.sig.return_annotation

    @return_annotation.setter
    def return_annotation(self, val: Any) -> None:
        self.sig = self.sig.replace(return_annotation=val)

    @return_annotation.deleter
    def return_annotation(self) -> None:
        self.sig = self.sig.replace(
            return_annotation=self.original.return_annotation
        )

    def add_parameter(
        self,
        name: str,
        kind: str = "POSITIONAL_OR_KEYWORD",
        default: Any = EMPTY,
        annotation: Any = EMPTY
    ) -> None:
        """Add a parameter to the signature.

        Parameters
        ----------
        name : str
            The name of the parameter to add.
        kind : str, default "POSITIONAL_OR_KEYWORD"
            A string specifying the species of parameter to add.  Must be one
            of the kinds specified in
            :class:`inspect.Parameter.kind <python:inspect.Parameter.kind>`.
        default : Any, default EMPTY
            The default value for the parameter.  If the parameter has no
            default, leave this empty.
        annotation : Any, default EMPTY
            The type annotation for the parameter.  If the parameter has no
            annotation, leave this empty.

        Raises
        ------
        ValueError
            If the named argument is already present in the signature, or if
            ``kind`` is not one of the kinds specified in
            :class:`inspect.Parameter.kind <python:inspect.Parameter.kind>`.

        See Also
        --------
        Signature.set_parameter :
            Modify a parameter in the signature.
        Signature.delete_parameter :
            Remove a parameter from the signature.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.add_parameter("qux", kind="KEYWORD_ONLY", default=3, annotation=int)
            >>> sig
            <Signature (bar, baz=2, *, qux: int = 3, **kwargs)>
            >>> sig.add_parameter("corge", kind="POSITIONAL_ONLY")
            >>> sig
            <Signature (corge, /, bar, baz=2, *, qux: int = 3, **kwargs)>
            >>> sig.add_parameter("args", kind="VAR_POSITIONAL")
            >>> sig
            <Signature (corge, /, bar, baz=2, *args, qux: int = 3, **kwargs)>
        """
        if name in self.parameter_map:
            raise ValueError(f"Parameter {name} already exists")

        try:
            par_kind = KINDS[kind]
        except KeyError as err:
            raise ValueError(f"Invalid kind: {repr(kind)}") from err

        # construct Parameter
        parameter = inspect.Parameter(
            name=name,
            kind=par_kind,
            default=default,
            annotation=annotation
        )

        # respect Python argument order based on kind.
        # NOTE: we count from the right to maintain lexicographic order.
        reverse = reversed(list(enumerate(self.parameters)))
        generator = (idx + 1 for idx, par in reverse if par.kind <= par_kind)
        index = next(generator, 0)

        # insert into parameters
        self.parameters = (
            self.parameters[:index] + (parameter,) + self.parameters[index:]
        )

    def set_parameter(
        self,
        _name: str,
        kind: str | None = None,
        **kwargs
    ) -> None:
        """Modify a parameter by name.

        Parameters
        ----------
        name : str
            The name of the parameter to modify.
        **kwargs
            Keyword arguments to pass to
            :meth:`Parameter.replace() <python:inspect.Parameter.replace>`.

        Raises
        ------
        KeyError
            If the named argument is not contained within the signature.

        See Also
        --------
        Signature.add_parameter :
            Add a parameter to the signature.
        Signature.delete_parameter :
            Remove a parameter from the signature.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.set_parameter("bar", annotation=int)
            >>> sig
            <Signature (bar: int, baz=2, **kwargs)>
            >>> sig.set_parameter("baz", default=3)
            >>> sig
            <Signature (bar: int, baz=3, **kwargs)>
            >>> sig.set_parameter("baz", name="qux")
            >>> sig
            <Signature (bar: int, qux=3, **kwargs)>
            >>> sig.set_parameter("bar", kind="KEYWORD_ONLY")
            >>> sig
            <Signature (qux=3, *, bar: int, **kwargs)>
        """
        if _name not in self.parameter_map:
            raise KeyError(_name)

        # replace parameter in-place
        if kind is None:
            self.parameters = tuple(
                par.replace(**kwargs) if par.name == _name else par
                for par in self.parameters
            )

        # reorder the signature to respect the new kind
        else:
            par_kind = KINDS[kind]
            param = self.parameter_map[_name].replace(kind=par_kind, **kwargs)

            # remove parameter
            index = self.parameter_index(_name)
            parameters = (
                self.parameters[:index] + self.parameters[index + 1:]
            )

            # add parameter
            reverse = reversed(list(enumerate(parameters)))
            gen = (idx + 1 for idx, par in reverse if par.kind <= par_kind)
            index = next(gen, 0)
            self.parameters = (
                parameters[:index] + (param,) + parameters[index:]
            )

    def reset_parameter(self, name: str) -> None:
        """Replace a parameter with its original definition, if one exists.

        Parameters
        ----------
        name : str
            The name of the parameter to replace.

        Raises
        ------
        KeyError
            If the named argument is not a parameter of this signature.
        TypeError
            If the named parameter does not have an original definition.

        See Also
        --------
        Signature.add_parameter :
            Add a parameter to the signature.
        Signature.set_parameter :
            Modify a parameter in the signature.
        Signature.delete_parameter :
            Remove a parameter from the signature.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.add_parameter("qux", default=3)
            >>> sig.set_parameter("bar", annotation=int, default=3)
            >>> sig
            <Signature (bar: int = 1, baz=2, qux=3, **kwargs)>
            >>> sig.reset_parameter("bar")
            >>> sig
            <Signature (bar, baz=2, qux=3, **kwargs)>
            >>> sig.reset_parameter("qux")
            Traceback (most recent call last):
                ...
            TypeError: 'qux' does not appear in the original signature
        """
        # get index of named parameter
        index = self.parameter_index(name)

        # check if parameter exists in original signature
        if name not in self.original.parameters:
            raise TypeError(
                f"'{name}' does not appear in the original signature"
            )

        # replace parameter
        self.parameters = (
            self.parameters[:index] +
            (self.original.parameters[name], ) +
            self.parameters[index + 1:]
        )

    def delete_parameter(self, name: str) -> None:
        """Remove a parameter from this signature.

        Parameters
        ----------
        name : str
            The name of the parameter to remove.

        Raises
        ------
        KeyError
            If the named argument is not a parameter of this signature.

        See Also
        --------
        Signature.add_parameter :
            Add a parameter to the signature.
        Signature.set_parameter :
            Modify a parameter in the signature.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.delete_parameter("bar")
            >>> sig
            <Signature (baz=2, **kwargs)>
            >>> sig.delete_parameter("kwargs")
            >>> sig
            <Signature (baz=2)>
            >>> sig.delete_parameter("baz")
            >>> sig
            <Signature ()>
        """
        # get index of named parameter
        index = self.parameter_index(name)

        # drop parameter
        self.parameters = self.parameters[:index] + self.parameters[index + 1:]

    def parameter_index(self, name: str) -> int:
        """Find the index of a parameter by name.

        Parameters
        ----------
        name : str
            The name of the parameter to find.

        Returns
        -------
        int
            The index of the parameter.

        Raises
        ------
        KeyError
            If the named argument is not a parameter of this signature.

        See Also
        --------
        Signature.move_parameter :
            Move a parameter to a new index.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.parameter_index("baz")
            1
            >>> sig.parameter_index("kwargs")
            2
        """
        param = self.parameter_map[name]
        return self.parameters.index(param)

    def move_parameter(self, name: str, index: int) -> None:
        """Move a parameter to the specified index.

        Parameters
        ----------
        name : str
            The name of the parameter to move.
        index : int
            The index to move the parameter to.  This supports negative
            indexing just like a normal tuple.

        Raises
        ------
        KeyError
            If the named argument is not contained within the signature.
        IndexError
            If the index is out of bounds.
        ValueError
            If the new parameter ordering is invalid.  This can occur if, for
            instance, a keyword-only parameter is moved before a positional
            one.

        See Also
        --------
        Signature.parameter_index :
            Find the index of a parameter by name.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig
            <Signature (bar, baz, **kwargs)>
            >>> sig.move_parameter("baz", 0)
            >>> sig
            <Signature (baz, bar, **kwargs)>
            >>> sig.move_parameter("bar", -1)
            Traceback (most recent call last):
                ...
            ValueError: wrong parameter order: variadic keyword parameter before positional or keyword parameter
        """
        # treat negative indices as positive
        if not -len(self.parameters) <= index < len(self.parameters):
            raise IndexError("tuple index out of range")
        if index < 0:
            index += len(self.parameters)

        # get index of parameter
        curr_index = self.parameter_index(name)
        param = self.parameter_map[name]

        # remove parameter
        parameters = (
            self.parameters[:curr_index] + self.parameters[curr_index + 1:]
        )

        # insert parameter at specified index
        self.parameters = parameters[:index] + (param,) + parameters[index:]

    def reconstruct(
        self,
        defaults: bool = True,
        annotations: bool = True,
        return_annotation: bool = True,
    ) -> str:
        """Return a complete string representation of the signature.

        Parameters
        ----------
        annotations: bool, default True
            Indicates whether to include type annotations for each parameter in
            the resulting string.
        defaults: bool, default True
            Indicates whether to include default values for each parameter in
            the resulting string.
        return_annotation: bool, default True
            Indicates whether to include the return annotation in the resulting
            string.

        Returns
        -------
        str
            A string listing the function name and all its arguments, with or
            without type annotations/default values.

        Examples
        --------
        .. doctest::

            >>> def foo(bar: int, baz: int = 2, **kwargs) -> dict[str, int]:
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.reconstruct()
            'foo(bar: int, baz: int = 2, **kwargs) -> dict[str, int]'
            >>> sig.reconstruct(annotations=False)
            'foo(bar, baz=2, **kwargs) -> dict[str, int]'
            >>> sig.reconstruct(annotations=False, return_annotation=False)
            'foo(bar, baz=2, **kwargs)'
            >>> sig.reconstruct(defaults=False, annotations=False, return_annotation=False)
            'foo(bar, baz, **kwargs)'
        """
        # configure replacement kwargs
        kwargs: dict[str, type] = {}
        if not annotations:
            kwargs["annotation"] = EMPTY
        if not defaults:
            kwargs["default"] = EMPTY

        # replace parameters
        parameters = tuple(par.replace(**kwargs) for par in self.parameters)
        sig = self.sig.replace(parameters=parameters)

        # replace return annotation
        if not return_annotation:
            sig = sig.replace(return_annotation=EMPTY)

        # combine into string
        return f"{self.func_name}{sig}"

    def compatible(self, other: Signature) -> bool:
        """Return ``True`` if the other signature implements all of this
        function's arguments.

        Parameters
        ----------
        other : Signature
            The signature to match against.  This is free to define additional
            parameters beyond those of this signature, but it must implement
            at least these parameters for this method to return ``True``.

        Returns
        -------
        bool
            ``True`` if the other signature this signature's interface.
            ``False`` otherwise.

        Notes
        -----
        The following rules are used to determine compatibility:

            *   The other signature must implement every positional argument of
                this :class:`Signature <pdcast.Signature>` in the same order.
                They can have different annotations and/or default values, but
                they must have the same name and kind.  The other signature can
                also define additional positional arguments beyond those of
                this object, which are not compared.
            *   If this signature accepts variadic positional arguments, the
                other signature must also accept variadic positional arguments
                with the same name.
            *   The other signature must implement every keyword-only argument
                of this :class:`Signature <pdcast.Signature>`.  The order does
                not matter, and they can have different annotations and/or
                default values, but they must have the same name and kind.  The
                other signature can also define additional keyword-only
                arguments beyond those of this object, which are not compared.
            *   If this signature accepts variadic keyword arguments, the other
                signature must also accept variadic keyword arguments with the
                same name.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = Signature(foo)
            >>> sig.compatible(Signature(lambda bar, baz, **kwargs: None))
            True
            >>> sig.compatible(Signature(lambda bar, baz, *, qux=3, corge=4, **kwargs: None))
            True
            >>> sig.compatible(Signature(lambda x, y: None))
            False
        """
        positional = {KINDS["POSITIONAL_ONLY"], KINDS["POSITIONAL_OR_KEYWORD"]}
        keyword = {KINDS["KEYWORD_ONLY"], KINDS["VAR_KEYWORD"]}

        def compatible(par, other_par):
            return par.name == other_par.name and par.kind == other_par.kind

        # iterate through this signature's parameters
        idx = 0
        for par in self.parameters:
            # get corresponding parameter from other signature
            try:
                other_par = other.parameters[idx]
            except IndexError:
                return False

            # check all positional arguments match exactly
            if par.kind in positional and not compatible(par, other_par):
                return False

            # if signature accepts *args, skip remaining positional args
            elif par.kind == KINDS["VAR_POSITIONAL"]:
                while other_par.kind in positional:
                    idx += 1
                    try:
                        other_par = other.parameters[idx]
                    except IndexError:
                        return False

                # check next parameter matches exactly
                if not compatible(par, other_par):
                    return False

            # check keyword-only arguments are present (no specific order)
            elif par.kind in keyword:
                equivalent = other.parameter_map.get(par.name, None)
                if equivalent is None or not compatible(par, equivalent):
                    return False

            idx += 1

        return True

    def __call__(self, *args, **kwargs) -> Arguments:
        """Bind the arguments to the :class:`Signature <pdcast.Signature>`,
        returning a corresponding :class:`Arguments <pdcast.Arguments>` object.

        Parameters
        ----------
        *args, **kwargs
            Arbitrary positional and keyword arguments to bind to this
            signature.

        Returns
        -------
        Arguments
            A wrapper around a
            :class:`BoundArguments <python:inspect.BoundArguments>` object with
            additional context for decorator-related functionality.

        Notes
        -----
        The default implementation of this method is equivalent to a
        :meth:`bind_partial <python:inspect.Signature.bind_partial>` call on
        the underlying :class:`inspect.Signature <python:inspect..Signature>`
        object.
        
        Subclasses can override this method to provide additional functionality.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = pdcast.Signature(foo)
            >>> bound = sig(1)
            >>> bound
            <BoundArguments (bar=1)>
            >>> bound.apply_defaults()
            >>> bound
            <BoundArguments (bar=1, baz=2, kwargs={})>
            >>> bound.args, bound.kwargs
            ((1, 2), mappingproxy({}))
        """
        bound = self.sig.bind_partial(*args, **kwargs)
        return Arguments(bound=bound, signature=self)

    def __str__(self) -> str:
        return str(self.sig)

    def __repr__(self) -> str:
        return repr(self.sig)


class Arguments:
    """An extensible wrapper around a
    :class:`BoundArguments <python:inspect.BoundArguments>` object with extra
    context for decorator-related functionality.

    Parameters
    ----------
    bound : inspect.BoundArguments
        The :class:`BoundArguments <python:inspect.BoundArguments>` to wrap.
    signature : Signature
        A reference to the :class:`Signature <pdcast.Signature>` that spawned
        this object.

    Notes
    -----
    These are created by calling a :class:`Signature <pdcast.Signature>` using
    its :meth:`__call__() <pdcast.Signature.__call__>` method.

    Subclasses can add additional attributes to simplify decorator-specific
    operations.
    """

    def __init__(
        self,
        bound: inspect.BoundArguments,
        signature: Signature
    ):
        self.bound = bound
        self.signature = signature

    @property
    def values(self) -> dict[str, Any]:
        """A mutable mapping containing the current value of each argument.

        Returns
        -------
        dict[str, Any]
            An ordinary dictionary listing the bound value of each argument.
            Any changes that are made to this dictionary will be propagated
            to :attr:`args <pdcast.Arguments.args>` and
            :attr:`kwargs <pdcast.Arguments.kwargs>`.

        See Also
        --------
        Arguments.args :
            An ``*args`` tuple that can be used to invoke the original
            function.
        Arguments.kwargs :
            A ``**kwargs`` map that can be used to invoke the original
            function.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = pdcast.Signature(foo)
            >>> bound = sig(1)
            >>> bound.values
            {'bar': 1}
            >>> bound.apply_defaults()
            >>> bound.values
            {'bar': 1, 'baz': 2, 'kwargs': {}}
            >>> foo(*bound.args, **bound.kwargs)
            {'bar': 1, 'baz': 2}
            >>> bound.values["baz"] = 3
            >>> foo(*bound.args, **bound.kwargs)
            {'bar': 1, 'baz': 3}
        """
        return self.bound.arguments

    @property
    def args(self) -> tuple[Any, ...]:
        """Positional arguments that can be supplied to the original function.

        Returns
        -------
        tuple[Any, ...]
            An ``*args`` tuple that can be used to invoke the original
            function.

        See Also
        --------
        Arguments.values :
            A mutable dictionary containing the current value of each argument.
        Arguments.kwargs :
            A ``**kwargs`` map that can be used to invoke the original
            function.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = pdcast.Signature(foo)
            >>> bound = sig(1)
            >>> bound.args, bound.kwargs
            ((1,), mappingproxy({}))
            >>> bound.apply_defaults()
            >>> bound.args, bound.kwargs
            ((1, 2), mappingproxy({}))
        """
        return self.bound.args

    @property
    def kwargs(self) -> Mapping[str, Any]:
        """Keyword arguments that can be supplied to the original function.

        Returns
        -------
        Mapping[str, Any]
            A read-only ``**kwargs`` dictionary that can be used to invoke the
            original function.

        See Also
        --------
        Arguments.values :
            A mutable dictionary containing the current value of each argument.
        Arguments.args :
            An ``*args`` tuple that can be used to invoked the original
            function.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = pdcast.Signature(foo)
            >>> sig.add_parameter("qux", default=4)
            >>> sig
            <Signature (bar, baz=2, qux=4, **kwargs)>
            >>> bound = sig(1)
            >>> bound.args, bound.kwargs
            ((1,), mappingproxy({}))
            >>> bound.apply_defaults()
            >>> bound.args, bound.kwargs  # TODO: incorrect - probably need to do some custom shenanigans
            ((1, 2), mappingproxy({'qux': 4})
        """
        return MappingProxyType(self.bound.kwargs)

    def apply_defaults(self) -> None:
        """Apply the original signature's default values to the bound
        arguments.

        Examples
        --------
        .. doctest::

            >>> def foo(bar, baz=2, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> sig = pdcast.Signature(foo)
            >>> bound = sig(1)
            >>> bound
            <BoundArguments (bar=1)>
            >>> bound.apply_defaults()
            >>> bound
            <BoundArguments (bar=1, baz=2, kwargs={})>
        """
        return self.bound.apply_defaults()

    def __str__(self) -> str:
        return str(self.bound)

    def __repr__(self) -> str:
        return repr(self.bound)



"""This module defines the ``cast()`` function, as well as several stand-alone
equivalents that allow quick conversion to predefined data types.
"""
# pylint: disable=unused-argument
from __future__ import annotations
from typing import Any, Callable

import pandas as pd

from pdcast import types
from pdcast.decorators.attachable import attachable
from pdcast.decorators.base import FunctionDecorator
from pdcast.decorators.dispatch import dispatch
from pdcast.decorators.extension import extension_func
from pdcast.detect import detect_type
from pdcast.util.error import shorten_list
from pdcast.util.numeric import (
    downcast_integer, downcast_float, downcast_complex, within_tol
)
from pdcast.util.round import Tolerance
from pdcast.util.type_hints import type_specifier
from pdcast.util.vector import apply_with_errors

from . import arguments


# conversions
# +===========+===+===+===+===+===+===+===+===+===+
# |           | b | i | f | c | d | d | t | s | o |
# +===========+===+===+===+===+===+===+===+===+===+
# | bool      | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | int       | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | float     | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | complex   | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | decimal   | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | datetime  | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | timedelta | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | string    | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+
# | object    | x | x | x | x | x | x | x | x | x |
# +-----------+---+---+---+---+---+---+---+---+---+


# ignore this file when doing string-based object lookups in resolve_type()
IGNORE_FRAME_OBJECTS = True


class columnwise(FunctionDecorator):
    """A basic decorator that breaks up DataFrame inputs into individual
    columns before continuing with a conversion.

    Placing this above ``@extension_func`` allows validators to avoid
    implementing this themselves, and above ``@dispatch`` allows dispatched
    implementations to always work in one dimension.
    """
    # pylint: disable=invalid-name

    def __call__(
        self,
        data: Any,
        dtype: type_specifier | dict[str, type_specifier] = NotImplemented,
        **kwargs
    ):
        """Apply the wrapped function for each column independently."""
        # use default dtype if not given
        if dtype is NotImplemented:
            dtype = self.__wrapped__.dtype  # NOTE: from @extension_func

        # recursive DataFrame case
        if isinstance(data, pd.DataFrame):
            # broadcast across columns
            columns = data.columns
            if isinstance(dtype, dict):
                bad = [col for col in dtype if col not in columns]
                if bad:
                    raise ValueError(f"column not found: {repr(bad)}")
            else:
                dtype = dict.fromkeys(columns, dtype)

            # pass each column individually
            result = data.copy()
            for col, typespec in dtype.items():
                result[col] = self.__wrapped__(
                    result[col],
                    typespec,
                    **kwargs
                )
            return result

        # base case
        return self.__wrapped__(data, dtype, **kwargs)


class catch_errors(FunctionDecorator):
    """A basic decorator that enforces the ``errors="ignore"`` rule during
    conversions.

    Placing this above ``@dispatch`` allows dispatched implementations to
    avoid implementing this themselves.
    """
    # pylint: disable=invalid-name

    def __call__(
        self,
        data: Any,
        dtype: types.VectorType,
        *args,
        errors: str = NotImplemented,
        **kwargs
    ):
        """Call the wrapped function in a try/except block."""
        if errors is NotImplemented:
            errors = self.__wrapped__.errors  # NOTE: from @extension_func

        try:
            return self.__wrapped__(
                data,
                dtype,
                *args,
                errors=errors,
                **kwargs
            )

        # never ignore these errors
        except (KeyboardInterrupt, MemoryError, SystemError, SystemExit):
            raise

        # process according to `errors` arg
        except Exception as err:
            if errors == "ignore":
                return data
            raise err


######################
####    PUBLIC    ####
######################


@attachable
@columnwise
@extension_func
@catch_errors
@dispatch("series", "dtype", cache_size=128, rectify=True)
def cast(
    series: Any,
    dtype: type_specifier | None = None,
    **unused: Any,
) -> pd.Series | pd.DataFrame:
    """Cast arbitrary data to the specified data type.

    Parameters
    ----------
    data : Any
        The data to convert.  This can be a scalar or 1D iterable containing
        arbitrary data.
    dtype : type specifier
        The target :doc:`type </content/types/types>` for this conversion.
        This can be in any format recognized by :func:`resolve_type`.  It can
        also be omitted to perform an :ref:`anonymous <cast.anonymous>`
        conversion.
    **kwargs : dict
        Arbitrary keyword :ref:`arguments <cast.arguments>` used to customize
        the conversion.

    Notes
    -----
    This function uses `multiple dispatch <https://en.wikipedia.org/wiki/Multiple_dispatch>`
    
    This function dispatches to one of the
    :ref:`delegated <atomic_type.conversions>` conversion methods that are
    attached to each :class:`ScalarType`.  Types can override these methods to
    change the behavior of :func:`cast`.  The method that is chosen is based on
    the :attr:`family <ScalarType.family>` of its ``dtype`` argument.
    """
    series_type = detect_type(series)

    # recursively unwrap decorators and retry
    # NOTE: These operate like recursive stacks.  If decorators are detected,
    # they are popped off the stack in FIFO order before recurring.  We do this
    # first to the data, and then to the target data type, allowing conversions
    # to safely ignore decorators in all their implementations.

    for before in getattr(series_type, "decorators", ()):
        series = before.inverse_transform(series)
        return cast(series, dtype, **unused)

    for before in getattr(dtype, "decorators", ()):
        series = cast(series, dtype.wrapped, **unused)
        return before.transform(series)

    raise NotImplementedError(
        f"no conversion found between {str(series_type)} and {str(dtype)}"
    )


def to_boolean(
    data: Any,
    dtype: type_specifier = "bool",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to boolean data types."""
    dtype = arguments.dtype(dtype, {}, supertype="bool")
    return cast(data, dtype, **kwargs)


def to_integer(
    data: Any,
    dtype: type_specifier = "int",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to integer data types."""
    dtype = arguments.dtype(dtype, {}, supertype="int")
    return cast(data, dtype, **kwargs)


def to_float(
    data: Any,
    dtype: type_specifier = "float",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to floating point data types."""
    dtype = arguments.dtype(dtype, {}, supertype="float")
    return cast(data, dtype, **kwargs)


def to_complex(
    data: Any,
    dtype: type_specifier = "complex",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to complex data types."""
    dtype = arguments.dtype(dtype, {}, supertype="complex")
    return cast(data, dtype, **kwargs)


def to_decimal(
    data: Any,
    dtype: type_specifier = "decimal",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to decimal data types."""
    dtype = arguments.dtype(dtype, {}, supertype="decimal")
    return cast(data, dtype, **kwargs)


def to_datetime(
    data: Any,
    dtype: type_specifier = "datetime",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to datetime data types."""
    dtype = arguments.dtype(dtype, {}, supertype="datetime")
    return cast(data, dtype, **kwargs)


def to_timedelta(
    data: Any,
    dtype: type_specifier = "timedelta",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to timedelta data types."""
    dtype = arguments.dtype(dtype, {}, supertype="timedelta")
    return cast(data, dtype, **kwargs)


def to_string(
    data: Any,
    dtype: type_specifier = "string",
    **kwargs
) -> pd.Series:
    """Stand-alone conversion to string data types."""
    dtype = arguments.dtype(dtype, {}, supertype="string")
    return cast(data, dtype, **kwargs)


#######################
####    GENERIC    ####
#######################


# NOTE: the following are base conversions for each of the major data types.
# They are used if no specific implementation is given for the observed data.


wildcard = types.registry.roots


@cast.overload(wildcard, "bool")
def generic_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert a to boolean representation."""
    # trivial case
    if detect_type(series) == dtype:
        return series

    target = dtype.dtype

    # NOTE: pandas complains about converting arbitrary objects to nullable
    # boolean.  Luckily, NAs are dropped, so we can do a two-step conversion.

    if isinstance(target, types.ObjectDtype):
        series = apply_with_errors(series, dtype.type_def, errors=errors)
    elif target == pd.BooleanDtype():
        series = series.astype(target.numpy_dtype, copy=False)
    return series.astype(target)


@cast.overload(wildcard, "int")
def generic_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    errors: str,
    downcast: types.CompositeType,
    **unused
) -> pd.Series:
    """Convert arbitrary data to integer representation."""
    if detect_type(series) != dtype:  # ignore trivial
        target = dtype.dtype

        # NOTE: pandas complains about converting arbitrary objects to nullable
        # int. Luckily, NAs are dropped, so we can do a two-step conversion.

        if isinstance(target, types.ObjectDtype):
            series = apply_with_errors(series, dtype.type_def, errors=errors)
        elif target in pandas_integer_dtypes:
            series = series.astype(target.numpy_dtype, copy=False)
        series = series.astype(target)

    # downcast if applicable
    if downcast is not None:
        return downcast_integer(series, tol=tol, smallest=downcast)
    return series


@cast.overload(wildcard, "float")
def generic_to_float(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert arbitrary data to float representation."""
    if detect_type(series) != dtype:  # ignore trivial
        target = dtype.dtype
        if isinstance(target, types.ObjectDtype):
            series = apply_with_errors(series, dtype.type_def, errors=errors)
        series = series.astype(target)

    if downcast is not None:
        return downcast_float(series, tol=tol, smallest=downcast)
    return series


@cast.overload(wildcard, "complex")
def generic_to_complex(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert arbitrary data to complex representation."""
    if detect_type(series) != dtype:  # ignore trivial
        target = dtype.dtype
        if isinstance(target, types.ObjectDtype):
            series = apply_with_errors(series, dtype.type_def, errors=errors)
        series = series.astype(target)

    if downcast is not None:
        return downcast_complex(series, tol=tol, smallest=downcast)
    return series


@cast.overload(wildcard, "decimal")
def generic_to_decimal(
    series: pd.Series,
    dtype: types.ScalarType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert arbitrary data to decimal representation."""
    # trivial case
    if detect_type(series) == dtype:
        return series

    target = dtype.dtype
    if isinstance(target, types.ObjectDtype):
        series = apply_with_errors(series, dtype.type_def, errors=errors)
    return series.astype(target)


@cast.overload(wildcard, "datetime")
def generic_to_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    **unused
) -> pd.Series:
    """Convert arbitrary data to datetime representation."""
    # 2-step conversion: X -> decimal, decimal -> datetime
    series = cast(series, "decimal", errors="raise")
    return cast(series, dtype, **unused)


@cast.overload(wildcard, "timedelta")
def generic_to_timedelta(
    series: pd.Series,
    dtype: types.ScalarType,
    **unused
) -> pd.Series:
    """Convert arbitrary data to timedelta representation."""
    # 2-step conversion: X -> decimal, decimal -> timedelta
    series = cast(series, "decimal", errors="raise")
    return cast(series, dtype=dtype, **unused)


@cast.overload(wildcard, "string")
def generic_to_string(
    series: pd.Series,
    dtype: types.ScalarType,
    format: str,
    errors: str,
    **unused
) -> pd.Series:
    """Convert arbitrary data to string representation."""
    if detect_type(series) == dtype:
        return series

    if format:
        call = lambda x: f"{x:{format}}"
        series = apply_with_errors(series, call, errors=errors)

    target = dtype.dtype
    if isinstance(target, types.ObjectDtype):
        series = apply_with_errors(series, dtype.type_def, errors=errors)
    return series.astype(target, copy=False)


@cast.overload(wildcard, "object")
def generic_to_object(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str,
    **unused
) -> pd.Series:
    """Convert arbitrary data to string representation."""
    # trivial case
    if detect_type(series) == dtype:
        return series

    if call is None:
        call = dtype.type_def

    # trivial case: object root type
    if call is object:
        return series.astype(object, copy=False)

    return safe_apply(series=series, dtype=dtype, call=call, errors=errors)


#######################
####    PRIVATE    ####
#######################


pandas_integer_dtypes = {
    pd.Int8Dtype(), pd.Int16Dtype(), pd.Int32Dtype(), pd.Int64Dtype(),
    pd.UInt8Dtype(), pd.UInt16Dtype(), pd.UInt32Dtype(), pd.UInt64Dtype()
}


def safe_apply(
    series: pd.Series,
    dtype: types.ScalarType,
    call: Callable,
    errors: str
) -> pd.Series:
    """Apply a callable over the input series that produces objects of the
    appropriate type.

    This is only used for conversions to or from
    :class:`ObjectTypes <pdcast.ObjectType>`.
    """
    def safe_call(val):
        result = call(val)
        output_type = type(result)
        if output_type != dtype.type_def:
            raise TypeError(
                f"`call` must return an object of type {dtype.type_def}"
            )
        return result

    # apply `safe_call` over series and pass to delegated conversion
    series = apply_with_errors(series, safe_call, errors=errors)
    return series.astype(dtype.dtype)


def snap_round(
    series: pd.Series,
    tol: Tolerance,
    rule: str | None,
    errors: str
) -> pd.Series:
    """Snap a series to the nearest integer within `tol`, and then round
    any remaining results according to the given rule.  Reject any outputs
    that are not integer-like by the end of this process.
    """
    from pdcast.patch.round import round as round_generic

    # if rounding to nearest, don't bother applying tolerance
    nearest = (
        "half_floor", "half_ceiling", "half_down", "half_up", "half_even"
    )
    if rule in nearest:
        return round_generic(series, rule=rule)

    # apply tolerance & check for non-integer results
    if tol or rule is None:
        rounded = round_generic(series, rule="half_even")
        outside = ~within_tol(series, rounded, tol=tol)
        if tol:
            series = series.where(outside, rounded)

        # check for non-integer (ignore if rounding in next step)
        if rule is None and outside.any():
            if errors == "coerce":
                series = round_generic(series, "down")  # emulate int()
            else:
                raise ValueError(
                    f"precision loss exceeds tolerance {float(tol):g} at "
                    f"index {shorten_list(outside[outside].index.values)}"
                )

    # apply final rounding
    if rule:
        series = round_generic(series, rule=rule)

    return series



"""Mypy stubs for pdcast/util/string.pyx"""

# TODO: deprecate this in function in favor of full regex matching
def boolean_match(
    val: str, lookup: dict[str, int], ignore_case: bool, fill: int
) -> int:
    ...

def int_to_base(val: int, base: int, width: int) -> str: ...



"""Mypy stubs for pdcast/structs/linked.cpp"""
from typing import (
    Any, Callable, ClassVar, Generic, Iterable, Iterator, Mapping, TypeVar
)


K = TypeVar("K", covariant=True)
V = TypeVar("V", covariant=True)


class Lock:

    class Guard:
        
        @property
        def active(self) -> bool: ...
        @property
        def is_shared(self) -> bool: ...
        def __enter__(self) -> Lock.Guard: ...
        def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...

    # basic locks
    def __call__(self) -> Lock.Guard: ...  # exclusive
    def shared(self) -> Lock.Guard: ...  # shared

    # diagnostic locks
    def count(self) -> int: ...
    def duration(self) -> float: ...
    def contention(self) -> float: ...
    def reset_diagnostics(self) -> None: ...


class MemGuard:

    @property
    def active(self) -> bool: ...
    def __enter__(self) -> MemGuard: ...
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...


##########################
####    LINKEDLIST    ####
##########################


class LinkedList(Generic[V]):

    def __init__(
        self,
        values: Iterable[V] | None = None,
        max_size: int | None = None,
        spec: Any = None,
        reverse: bool = False,
        singly_linked: bool = False
    ) -> None:
        ...

    ####################
    ####    BASE    ####
    ####################

    __hash__: ClassVar[None] = ...  # type: ignore

    @property
    def SINGLY_LINKED(self) -> bool: ...
    @property
    def DOUBLY_LINKED(self) -> bool: ...
    @property
    def FIXED_SIZE(self) -> bool: ...
    @property
    def DYNAMIC(self) -> bool: ...
    @property
    def STRICTLY_TYPED(self) -> bool: ...
    @property
    def LOOSELY_TYPED(self) -> bool: ...
    @property
    def lock(self) -> Lock: ...
    @property
    def capacity(self) -> int: ...
    @property
    def max_size(self) -> int: ...
    @property
    def frozen(self) -> bool: ...
    @property
    def specialization(self) -> Any: ...
    @property
    def nbytes(self) -> int: ...

    def reserve(self, capacity: int) -> MemGuard: ...
    def defragment(self) -> None: ...
    def specialize(self, spec: Any) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[V]: ...
    def __reversed__(self) -> Iterator[V]: ...

    ####################
    ####    LIST    ####
    ####################

    def append(self, value: V) -> None: ...
    def append_left(self, value: V) -> None: ...
    def insert(self, index: int, value: V) -> None: ...
    def extend(self, values: Iterable[V]) -> None: ...
    def extend_left(self, values: Iterable[V]) -> None: ...
    def pop(self, index: int = -1) -> V: ...
    def remove(self, value: V) -> None: ...
    def clear(self) -> None: ...
    def copy(self) -> LinkedList[V]: ...
    def sort(
        self,
        *,
        key: Callable[[V], Any] | None = None,
        reverse: bool = False
    ) -> None:
        ...
    def reverse(self) -> None: ...
    def rotate(self, shift: int = 1) -> None: ...
    def count(self, value: V) -> int: ...
    def index(self, value: V, start: int = 0, stop: int = -1) -> int: ...
    def __contains__(self, item: V) -> bool: ...
    def __getitem__(self, index: int | slice) -> V | LinkedList[V]: ...
    def __setitem__(self, index: int | slice, value: V | Iterable[V]) -> None: ...
    def __delitem__(self, index: int | slice) -> None: ...
    def __add__(self, other: Iterable[V]) -> LinkedList[V]: ...
    def __iadd__(self, other: Iterable[V]) -> LinkedList[V]: ...
    def __mul__(self, repeat: int) -> LinkedList[V]: ...
    def __imul__(self, repeat: int) -> LinkedList[V]: ...
    def __lt__(self, other: Iterable[V]) -> bool: ...
    def __le__(self, other: Iterable[V]) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __gt__(self, other: Iterable[V]) -> bool: ...
    def __ge__(self, other: Iterable[V]) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


#########################
####    LINKEDSET    ####
#########################


class LinkedSet(Generic[K]):

    def __init__(
        self,
        values: Iterable[K] | None = None,
        max_size: int | None = None,
        spec: Any = None,
        reverse: bool = False,
        singly_linked: bool = False
    ) -> None:
        ...

    ####################
    ####    BASE    ####
    ####################

    __hash__: ClassVar[None] = ...  # type: ignore

    @property
    def SINGLY_LINKED(self) -> bool: ...
    @property
    def DOUBLY_LINKED(self) -> bool: ...
    @property
    def FIXED_SIZE(self) -> bool: ...
    @property
    def DYNAMIC(self) -> bool: ...
    @property
    def STRICTLY_TYPED(self) -> bool: ...
    @property
    def LOOSELY_TYPED(self) -> bool: ...
    @property
    def lock(self) -> Lock: ...
    @property
    def capacity(self) -> int: ...
    @property
    def max_size(self) -> int: ...
    @property
    def frozen(self) -> bool: ...
    @property
    def specialization(self) -> Any: ...
    @property
    def nbytes(self) -> int: ...

    def reserve(self, capacity: int) -> MemGuard: ...
    def defragment(self) -> None: ...
    def specialize(self, spec: Any) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[K]: ...
    def __reversed__(self) -> Iterator[K]: ...

    ####################
    ####    LIST    ####
    ####################

    def insert(self, index: int, value: K) -> None: ...
    def pop(self, index: int = -1) -> K: ...
    def clear(self) -> None: ...
    def copy(self) -> LinkedSet[K]: ...
    def sort(
        self,
        *,
        key: Callable[[K], Any] | None = None,
        reverse: bool = False
    ) -> None:
        ...
    def reverse(self) -> None: ...
    def rotate(self, shift: int = 1) -> None: ...
    def count(self, value: K) -> int: ...
    def index(self, value: K, start: int = 0, stop: int = -1) -> int: ...
    def __getitem__(self, index: int | slice) -> K | LinkedSet[K]: ...
    def __setitem__(self, index: int | slice, value: K | Iterable[K]) -> None: ...
    def __delitem__(self, index: int | slice) -> None: ...

    ###################
    ####    SET    ####
    ###################

    def add(self, value: K) -> None: ...
    def add_left(self, value: K) -> None: ...
    def lru_add(self, value: K) -> None: ...
    def remove(self, value: K) -> None: ...
    def discard(self, value: K) -> None: ...
    def union(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def union_left(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def intersection(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def difference(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def symmetric_difference(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def update(self, other: Iterable[K]) -> None: ...
    def update_left(self, other: Iterable[K]) -> None: ...
    def lru_update(self, other: Iterable[K]) -> None: ...
    def intersection_update(self, other: Iterable[K]) -> None: ...
    def difference_update(self, other: Iterable[K]) -> None: ...
    def symmetric_difference_update(self, other: Iterable[K]) -> None: ...
    def symmetric_difference_update_left(self, other: Iterable[K]) -> None: ...
    def isdisjoint(self, other: Iterable[K]) -> bool: ...
    def issubset(self, other: Iterable[K]) -> bool: ...
    def issuperset(self, other: Iterable[K]) -> bool: ...
    def distance(self, value1: K, value2: K) -> int: ...
    def swap(self, value1: K, value2: K) -> None: ...
    def move(self, value: K, shift: int) -> None: ...
    def move_to_index(self, value: K, index: int) -> None: ...
    def lru_contains(self, value: K) -> bool: ...
    def __contains__(self, item: K) -> bool: ...
    def __or__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __and__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __sub__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __xor__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __ior__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __iand__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __isub__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __ixor__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __lt__(self, other: Iterable[K]) -> bool: ...
    def __le__(self, other: Iterable[K]) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __gt__(self, other: Iterable[K]) -> bool: ...
    def __ge__(self, other: Iterable[K]) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


##########################
####    LINKEDDICT    ####
##########################


class KeysProxy(Generic[K, V]):

    @property
    def mapping(self) -> Mapping[K, V]: ...

    def index(self, key: K) -> int: ...
    def count(self, key: K) -> int: ...
    def isdisjoint(self, other: Iterable[K]) -> bool: ...
    def issubset(self, other: Iterable[K]) -> bool: ...
    def issuperset(self, other: Iterable[K]) -> bool: ...
    def union(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def union_left(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def intersection(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def difference(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def symmetric_difference(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def symmetric_difference_left(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __contains__(self, key: K) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[K]: ...
    def __reversed__(self) -> Iterator[K]: ...
    def __getitem__(self, index: int | slice) -> K | KeysProxy[K, V]: ...
    def __or__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __and__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __sub__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __xor__(self, other: Iterable[K]) -> LinkedSet[K]: ...
    def __lt__(self, other: Iterable[K]) -> bool: ...
    def __le__(self, other: Iterable[K]) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __gt__(self, other: Iterable[K]) -> bool: ...
    def __ge__(self, other: Iterable[K]) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


class ValuesProxy(Generic[K, V]):

    @property
    def mapping(self) -> Mapping[K, V]: ...

    def index(self, value: V) -> int: ...
    def count(self, value: V) -> int: ...
    def __contains__(self, value: V) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[V]: ...
    def __reversed__(self) -> Iterator[V]: ...
    def __getitem__(self, index: int | slice) -> V | ValuesProxy[K, V]: ...
    def __add__(self, other: Iterable[Any]) -> LinkedList[V]: ...
    def __mul__(self, repeat: int) -> LinkedList[V]: ...
    def __lt__(self, other: Iterable[V]) -> bool: ...
    def __le__(self, other: Iterable[V]) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __gt__(self, other: Iterable[V]) -> bool: ...
    def __ge__(self, other: Iterable[V]) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


class ItemsProxy(Generic[K, V]):

    @property
    def mapping(self) -> Mapping[K, V]: ...

    def index(self, item: tuple[K, V]) -> int: ...
    def count(self, item: tuple[K, V]) -> int: ...
    def __contains__(self, item: tuple[K, V]) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[tuple[K, V]]: ...
    def __reversed__(self) -> Iterator[tuple[K, V]]: ...
    def __getitem__(self, index: int | slice) -> tuple[K, V] | ItemsProxy[K, V]: ...
    def __add__(self, other: Iterable[Any]) -> LinkedList[tuple[K, V]]: ...
    def __mul__(self, repeat: int) -> LinkedList[tuple[K, V]]: ...
    def __lt__(self, other: Iterable[tuple[K, V]]) -> bool: ...
    def __le__(self, other: Iterable[tuple[K, V]]) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __gt__(self, other: Iterable[tuple[K, V]]) -> bool: ...
    def __ge__(self, other: Iterable[tuple[K, V]]) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...


class LinkedDict(Generic[K, V]):

    def __init__(
        self,
        items: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]] | None = None,
        max_size: int | None = None,
        spec: Any = None,
        reverse: bool = False,
        singly_linked: bool = False
    ) -> None:
        ...

    @classmethod
    def fromkeys(
        cls,
        keys: Iterable[K], value: V | None = None,
        max_size: int | None = None,
        spec: Any = None,
        singly_linked: bool = False
    ) -> LinkedDict[K, V]:
        ...

    ####################
    ####    BASE    ####
    ####################

    __hash__: ClassVar[None] = ...  # type: ignore

    @property
    def SINGLY_LINKED(self) -> bool: ...
    @property
    def DOUBLY_LINKED(self) -> bool: ...
    @property
    def FIXED_SIZE(self) -> bool: ...
    @property
    def DYNAMIC(self) -> bool: ...
    @property
    def STRICTLY_TYPED(self) -> bool: ...
    @property
    def LOOSELY_TYPED(self) -> bool: ...
    @property
    def lock(self) -> Lock: ...
    @property
    def capacity(self) -> int: ...
    @property
    def max_size(self) -> int: ...
    @property
    def frozen(self) -> bool: ...
    @property
    def specialization(self) -> Any: ...
    @property
    def nbytes(self) -> int: ...

    def reserve(self, capacity: int) -> MemGuard: ...
    def defragment(self) -> None: ...
    def specialize(self, spec: Any) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[K]: ...
    def __reversed__(self) -> Iterator[K]: ...

    ####################
    ####    LIST    ####
    ####################

    def clear(self) -> None: ...
    def copy(self) -> LinkedDict[K, V]: ...
    def sort(
        self,
        *,
        key: Callable[[K, V], Any] | None = None,
        reverse: bool = False
    ) -> None:
        ...
    def reverse(self) -> None: ...
    def rotate(self, shift: int = 1) -> None: ...
    def count(self, key: K) -> int: ...
    def index(self, key: K, start: int = 0, stop: int = -1) -> int: ...

    ###################
    ####    SET    ####
    ###################

    def remove(self, key: K) -> None: ...
    def discard(self, key: K) -> None: ...
    def union(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> LinkedDict[K, V]:
        ...
    def union_left(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> LinkedDict[K, V]:
        ...
    def intersection(self, other: Iterable[K]) -> LinkedDict[K, V]: ...
    def difference(self, other: Iterable[K]) -> LinkedDict[K, V]: ...
    def symmetric_difference(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> LinkedDict[K, V]:
        ...
    def symmetric_difference_left(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> LinkedDict[K, V]:
        ...
    def update(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> None:
        ...
    def update_left(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> None:
        ...
    def lru_update(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> None:
        ...
    def intersection_update(self, other: Iterable[K]) -> None: ...
    def difference_update(self, other: Iterable[K]) -> None: ...
    def symmetric_difference_update(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> None:
        ...
    def symmetric_difference_update_left(
        self, other: dict[K, V] | LinkedDict[K, V] | Iterable[tuple[K, V]]
    ) -> None:
        ...

    ####################
    ####    DICT    ####
    ####################

    def add(self, key: K, value: V) -> None: ...
    def add_left(self, key: K, value: V) -> None: ...
    def lru_add(self, key: K, value: V) -> None: ...
    def insert(self, index: int, key: K, value: V) -> None: ...
    def pop(self, key: K, default: V | None = None) -> V: ...
    def popitem(self, index: int = -1) -> tuple[K, V]: ...
    def get(self, key: K, default: V | None = None) -> V: ...
    def lru_get(self, key: K, default: V | None = None) -> V: ...
    def setdefault(self, key: K, default: V | None = None) -> V: ...
    def setdefault_left(self, key: K, default: V | None = None) -> V: ...
    def lru_setdefault(self, key: K, default: V | None = None) -> V: ...
    def keys(self) -> KeysProxy[K, V]: ...
    def values(self) -> ValuesProxy[K, V]: ...
    def items(self) -> ItemsProxy[K, V]: ...
    def __getitem__(self, key: K) -> V: ...
    def __setitem__(self, key: K, value: V) -> None: ...
    def __delitem__(self, key: K) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...



"""This package provides utilities for converting to, from, and between various
datetime and timedelta representations.

Modules
-------
calendar
    Vectorized math for conversions involving the proleptic Gregorian calendar.

datetime
    Utilities for converting to and from various datetime representations.

epoch
    Customizable epochs for datetime/timedelta calculations.

timedelta
    Utilities for converting to and from various timedelta representations.

unit
    Datetime and timedelta unit conversions.
"""
from .calendar import (
    date_to_days, days_in_month, days_to_date, is_leap_year, leaps_between
)
from .datetime import (
    filter_dateutil_parser_error, is_iso_8601_format_string, iso_8601_to_ns,
    is_utc, localize_pydatetime_scalar, numpy_datetime64_to_ns,
    ns_to_pydatetime, pandas_timestamp_to_ns, pydatetime_to_ns,
    string_to_pydatetime
)
from .epoch import Epoch
from .epoch import epoch_aliases_public as epoch_aliases
from .timedelta import (
    numpy_timedelta64_to_ns, pandas_timedelta_to_ns, pytimedelta_to_ns,
    timedelta_string_to_ns
)
from .timezone import tz
from .unit import convert_unit, round_months_to_ns, round_years_to_ns
from .unit import as_ns_public as as_ns
from .unit import valid_units_public as valid_units



"""This module describes a ``SparseType`` object, which can be used to
dynamically wrap other types.
"""
from __future__ import annotations
from typing import Any

import pandas as pd

from .base import EMPTY, REGISTRY, Decorator, DecoratorMeta, Empty, TypeMeta, resolve


# TODO: SparseType works, but not in all cases.
# -> pd.NA disallows non-missing fill values
# -> Timestamps must be sparsified manually by converting to object and then
# to sparse
# -> Timedeltas just don't work at all.  astype() rejects pd.SparseDtype("m8")
# entirely.


# TODO:
# >>> Sparse["long[numpy] | long[pandas]"]
# Traceback (most recent call last):
#   File "<stdin>", line 1, in <module>
#   File "/home/eerkela/data/bertrand/bertrand/types/base/meta.py", line 4698, in __getitem__
#     return cls.__class_getitem__(params)
#            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#   File "/home/eerkela/data/bertrand/bertrand/types/base/meta.py", line 7118, in wrapper
#     return Union.from_types(LinkedSet(
#                             ^^^^^^^^^^
# TypeError: SparseDtype subtype must be a numpy dtype


class Sparse(Decorator, cache_size=256):
    """TODO"""

    aliases = {pd.SparseDtype, "sparse"}
    _is_empty: bool = False


    def __class_getitem__(
        cls,
        wrapped: TypeMeta | DecoratorMeta | None = None,
        fill_value: Any | Empty = EMPTY
    ) -> DecoratorMeta:
        if wrapped is None:
            raise NotImplementedError("TODO")  # TODO: should never occur

        is_empty = fill_value is EMPTY

        if isinstance(wrapped.unwrapped, DecoratorMeta):
            return cls.flyweight(wrapped, fill_value, dtype=None, _is_empty=is_empty)

        if is_empty:
            fill = wrapped.missing
        elif pd.isna(fill_value):  # type: ignore
            fill = fill_value
        else:
            fill = wrapped(fill_value)
            if len(fill) != 1:
                raise TypeError(f"fill_value must be a scalar, not {repr(fill_value)}")
            fill = fill[0]

        dtype = pd.SparseDtype(wrapped.dtype, fill)
        return cls.flyweight(wrapped, fill, dtype=dtype, _is_empty=is_empty)

    @classmethod
    def from_string(cls, wrapped: str, fill_value: str | Empty = EMPTY) -> DecoratorMeta:
        """Resolve a sparse specifier in the type specification mini-langauge.
        """
        return cls[wrapped, REGISTRY.na_strings.get(fill_value, fill_value)]  # type: ignore

    @classmethod
    def from_dtype(cls, dtype: pd.SparseDtype) -> DecoratorMeta:
        """Convert a pandas SparseDtype into a
        :class:`SparseType <pdcast.SparseType>` object.
        """
        return cls[resolve(dtype.subtype), dtype.fill_value]  # type: ignore

    @classmethod
    def replace(cls, *args: Any, **kwargs: Any) -> DecoratorMeta:
        """TODO"""
        if cls._is_empty and len(args) < 2 and "fill_value" not in kwargs:
            kwargs["fill_value"] = EMPTY
        return super().replace(*args, **kwargs)

    @classmethod
    def inverse_transform(cls, series: pd.Series[Any]) -> pd.Series[Any]:
        """Convert a sparse series into a dense format"""
        return series.sparse.to_dense()



"""Mypy stubs for pdcast/util/round/tolerance.pyx"""
from decimal import Decimal

from pdcast.util.type_hints import numeric


class Tolerance:
    real: Decimal
    imag: Decimal

    def __init__(self, tol: numeric) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __bool__(self) -> bool: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __complex__(self) -> complex: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...



"""This package contains all the internal machinery and basic building blocks that
comprise the bertrand type system.
"""
from .meta import (
    ALIAS, DEBUG, DTYPE, EMPTY, META, POINTER_SIZE, REGISTRY, TYPESPEC, Aliases,
    BaseMeta, Decorator, DecoratorMeta, DecoratorPrecedence, Edges, Empty, StructuredMeta,
    StructuredUnion, Type, TypeMeta, TypeRegistry, Union, UnionMeta, detect, resolve
)



cdef dict epoch_aliases

cdef class Epoch:

    cdef readonly:
        object offset  # offset from UTC in ns
        long int year
        unsigned char month
        unsigned char day
        unsigned short ordinal



# pylint: disable=redefined-outer-name, unused-argument
from __future__ import annotations
import datetime
import zoneinfo

import tzlocal


def tz(tz: str | datetime.tzinfo | None, state: dict) -> datetime.tzinfo:
    """Convert a time zone specifier into a ``datetime.tzinfo`` object."""
    # naive
    if tz is None:
        return None

    # trivial case
    if isinstance(tz, datetime.tzinfo):
        return tz

    # string
    if tz.lower() == "local":
        return zoneinfo.ZoneInfo(tzlocal.get_localzone_name())
    return zoneinfo.ZoneInfo(tz)



"""This module describes an ``ObjectType`` object, which can be used to
represent arbitrary Python data.
"""
import numpy as np

from .base import Type, TypeMeta
from .base.meta import get_from_calling_context, synthesize_dtype


class Object(Type, backend="object", cache_size=128):
    """Generic object type for arbitrary python values.
    """

    aliases = {
        object, np.dtype(object), "object", "obj", "O", "pyobject", "object_",
        "object0",
    }
    scalar = object
    dtype = np.dtype(object)

    def __class_getitem__(cls, scalar: type = object) -> TypeMeta:
        if not isinstance(scalar, type):
            raise TypeError(f"scalar must be a Python type, not {repr(scalar)}")

        result = cls.flyweight(scalar, scalar=scalar)
        result._fields["dtype"] = None  # workaround to enable synthesize_dtype()
        result._fields["dtype"] = synthesize_dtype(result)
        return result

    @classmethod
    def from_string(cls, scalar: str) -> TypeMeta:
        """Resolve a string in the type specification mini-language.

        Notes
        -----
        This method can resolve type definitions (and even dotted variables!)
        from the calling context by inspecting the stack frame.

        >>> resolve_type("object[int]")
        ObjectType(type_def=<class 'int'>)
        """
        return cls[get_from_calling_context(scalar, skip=1)]



"""A set of high-performance data structures based on linked lists.

Each data structure is implemented in pure C++ using the CPython API, and can
be used just like their python `list`, `set`, and `dict` counterparts.  They
have all the same methods and semantics, plus a few extras related to their
ordered nature and enhanced type/thread safety.  The same data structures are
also available in C++ with identical interfaces, so you can use them in other
extension modules and C++ code.  They are located under the `<bertrand.h>`
header and associated bertrand:: namespace, both of which are installed along
with the python module using CMake integrations.

See the bertrand API documentation for more details.
"""
from .linked import LinkedList, LinkedSet, LinkedDict
